# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test merging.
=============
"""
import numpy
import zarr

# pylint: enable=unused-import
from .. import _update_fs, merge_time_series, perform
from ... import sync
from ...tests import data
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster
from ...tests.fs import local_fs


class MyError(RuntimeError):
    """Custom error."""
    ...


class ThrowError(sync.Sync):
    """Throw an error when merging."""

    def __enter__(self) -> bool:
        raise MyError('This is an error')

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...


def test_update_fs(
        dask_client,  # pylint: disable=redefined-outer-name
        local_fs,  # pylint: disable=redefined-outer-name
):
    """Test the _update_fs function."""
    generator = data.create_test_dataset()
    ds = next(generator)

    partition_folder = local_fs.root.joinpath('partition_folder')

    zattrs = str(partition_folder.joinpath('.zattrs'))
    future = dask_client.submit(_update_fs, str(partition_folder),
                                dask_client.scatter(ds), local_fs.fs)
    dask_client.gather(future)
    assert local_fs.exists(zattrs)

    local_fs.fs.rm(str(partition_folder), recursive=True)
    assert not local_fs.exists(zattrs)
    seen_exception = False
    try:
        future = dask_client.submit(_update_fs, str(partition_folder),
                                    dask_client.scatter(ds), local_fs.fs,
                                    ThrowError())
        dask_client.gather(future)
    except MyError:
        seen_exception = True
    assert seen_exception
    assert not local_fs.exists(zattrs)


def test_perform(
        local_fs,  # pylint: disable=redefined-outer-name
        dask_client,  # pylint: disable=redefined-outer-name
):
    """Test the perform function."""
    generator = data.create_test_dataset()
    ds = next(generator)

    path = str(local_fs.root.joinpath('folder'))

    future = dask_client.submit(_update_fs, path, dask_client.scatter(ds),
                                local_fs.fs)
    dask_client.gather(future)

    future = dask_client.submit(perform, dask_client.scatter(ds), path, 'time',
                                local_fs.fs, 'time', merge_time_series)
    dask_client.gather(future)

    zgroup = zarr.open_consolidated(local_fs.get_mapper(path))
    assert numpy.all(zgroup['time'][...] == ds['time'].values)
    assert numpy.all(zgroup['var1'][...] == ds['var1'].values)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
I/O operations
==============
"""
from __future__ import annotations

from typing import Any, Iterable, Sequence
import collections
import json
import logging

import dask.array.core
import dask.base
from dask.delayed import Delayed as dask_Delayed
import dask.distributed
import dask.local
import fsspec
import numpy
import zarr

from . import dask_utils, dataset, meta, sync
from .fs_utils import join_path
from .type_hints import ArrayLike

#: Block size limit used with dask arrays. (128 MiB)
BLOCK_SIZE_LIMIT = 134217728

#: Name of the attribute storing the names of the dimensions of an array.
DIMENSIONS = '_ARRAY_DIMENSIONS'

#: Configuration file that describes the attributes of an array.
ZATTRS = '.zattrs'

#: Configuration file that describes the attributes of a group.
ZGROUP = '.zgroup'

#: Module logger.
_LOGGER = logging.getLogger(__name__)


def execute_transaction(
    client: dask.distributed.Client,
    synchronizer: sync.Sync,
    futures: Sequence[dask.distributed.Future | dask_Delayed],
) -> Any:
    """Execute a transaction in the collection.

    Args:
        client: The Dask client.
        synchronizer: The instance handling access to critical resources.
        futures: Lazy tasks to be done.

    Returns:
        The result of the transaction.
    """
    if not futures:
        return None
    awaitables = []
    try:
        with synchronizer:
            awaitables = client.compute(futures)
            return client.gather(awaitables)
    except:  # noqa: E722
        # Before throwing the exception, we wait until all future scheduled
        # ones finished.
        dask.distributed.wait(awaitables)
        dask.distributed.wait(futures)
        raise


def _to_zarr(array: dask.array.core.Array, mapper: fsspec.FSMap, path: str,
             **kwargs) -> None:
    """Write a Dask array to a Zarr dataset.

    Args:
        array: The array to write.
        mapper: The file system mapper.
        path: The path to the Zarr dataset.
        **kwargs: Keyword arguments to pass to :func:`zarr.create`.
    """
    chunks = [chunk[0] for chunk in array.chunks]
    target = zarr.create(
        shape=array.shape,
        chunks=chunks,  # type: ignore[arg-type]
        dtype=array.dtype,
        store=mapper,
        path=path,
        overwrite=True,
        write_empty_chunks=False,
        **kwargs)
    dask.array.core.store(array,
                          target,
                          flush=True,
                          lock=False,
                          compute=True,
                          scheduler=dask.local.get_sync,
                          return_stored=False)


def write_zattrs(
    dirname: str,
    variable: meta.Variable | dataset.Variable,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Write the attributes of a variable to a Zarr dataset.

    Args:
        dirname: The storage directory of the Zarr dataset.
        variable: The variable to process.
        fs: The file system on which the Zarr dataset is stored.
    """
    attrs = collections.OrderedDict(item.get_config()
                                    for item in variable.attrs)
    attrs[DIMENSIONS] = variable.dimensions
    with fs.open(join_path(dirname, variable.name, ZATTRS),
                 mode='w') as stream:
        json.dump(attrs, stream, indent=2)  # type: ignore[arg-type]


def write_zarr_variable(
    args: tuple[str, dataset.Variable],
    dirname: str,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Write a variable to a Zarr dataset.

    Args:
        args: The arguments to the function:
            - Name of the variable to write.
            - The variable to write.
        dirname: The target directory.
        fs: The file system on which the Zarr dataset is stored.
    """
    name, variable = args
    kwargs = {'filters': variable.filters}
    data = variable.array

    chunks = {ix: -1 for ix in range(variable.ndim)}
    data = data.rechunk(
        chunks,  # type: ignore[arg-type]
        block_size_limit=BLOCK_SIZE_LIMIT,
    )

    _to_zarr(array=data,
             mapper=fs.get_mapper(dirname),
             path=name,
             compressor=variable.compressor,
             fill_value=variable.fill_value,
             **kwargs)
    write_zattrs(dirname, variable, fs)


def _write_meta(
    ds: dataset.Dataset,
    dirname: str,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Write the metadata of a dataset to a Zarr dataset.

    Args:
        ds: The dataset to process.
        dirname: The storage directory of the Zarr dataset.
        fs: The file system on which the Zarr dataset is stored.
    """
    attrs = collections.OrderedDict(item.get_config() for item in ds.attrs)
    with fs.open(join_path(dirname, ZATTRS), mode='w') as stream:
        json.dump(attrs, stream, indent=2)  # type: ignore[arg-type]

    with fs.open(join_path(dirname, ZGROUP), mode='w') as stream:
        json.dump(
            {'zarr_format': 2},
            stream,  # type: ignore[arg-type]
            indent=2,
        )
    zarr.consolidate_metadata(fs.get_mapper(dirname))  # type: ignore[arg-type]
    fs.invalidate_cache(dirname)


def write_zarr_group(
    ds: dataset.Dataset,
    dirname: str,
    fs: fsspec.AbstractFileSystem,
    synchronizer: sync.Sync,
    parallel: bool = True,
) -> None:
    """Write a partition to a Zarr group.

    Args:
        ds: The dataset to write.
        dirname: The name of the partition.
        fs: The file system that the partition is stored on.
        synchronizer: The instance handling access to critical resources.
        parallel: Whether to write the variables in parallel using Dask.
    """
    futures = [
        dask_utils.simple_delayed(name, write_zarr_variable)((name, variable),
                                                             dirname, fs)
        for name, variable in ds.variables.items()
    ]
    if parallel:
        with dask.distributed.worker_client() as client:
            execute_transaction(client, synchronizer, futures)
    else:
        dask.base.compute(*futures, scheduler=dask.local.get_sync)
    _write_meta(ds, dirname, fs)


def open_zarr_array(array: zarr.Array, name: str) -> dataset.Variable:
    """Open a Zarr array as a Dask array.

    Args:
        array: The Zarr array to open.
        name: The name of the variable.

    Returns:
        The variable.
    """
    return dataset.Variable.from_zarr(array, name, DIMENSIONS)


def open_zarr_group(
        dirname,
        fs: fsspec.AbstractFileSystem,
        selected_variables: Iterable[str] | None = None) -> dataset.Dataset:
    """Open a Zarr group stored in a partition.

    Args:
        dirname: The name of the partition.
        fs: The file system that the partition is stored on.
        selected_variables: The list of variables to retain from the Zarr
            group. If None, all variables are selected.

    Returns:
        The zarr group stored in the partition.
    """
    _LOGGER.debug('Opening Zarr group %r', dirname)
    store: zarr.Group = zarr.open_consolidated(  # type: ignore[arg-type]
        fs.get_mapper(dirname), )
    # Ignore unknown variables to retain.
    selected_variables = set(selected_variables) & set(
        store) if selected_variables is not None else set(store)
    variables = [
        open_zarr_array(store[name], name)  # type: ignore[arg-type]
        for name in selected_variables
    ]

    return dataset.Dataset(
        variables=variables,
        attrs=tuple(dataset.Attribute(*item) for item in store.attrs.items()))


def update_zarr_array(
    dirname: str,
    array: ArrayLike,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Update a Zarr array.

    Args:
        dirname: The storage directory of the Zarr array..
        array: The data updated to write.
        fs: The file system that the Zarr array is stored on.
    """
    _LOGGER.debug('Updating Zarr array %r', dirname)
    store = zarr.open_array(fs.get_mapper(dirname), mode='a')

    if isinstance(array, dask.array.core.Array):
        array = array.compute()

    if isinstance(array,
                  numpy.ma.MaskedArray) and store.fill_value is not None:
        array = array.filled(store.fill_value)

    store[:] = array

    # Invalidate any cached directory information.
    fs.invalidate_cache(dirname)


def del_zarr_array(
    dirname: str,
    name: str,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Delete a variable from a Zarr dataset.

    Args:
        dirname: The name of the dataset.
        name: The name of the variable.
        fs: The file system that the dataset is stored on.
    """
    _LOGGER.debug('Deleting Zarr array %r', dirname)
    path = join_path(dirname, name)
    if fs.exists(path):
        fs.rm(path, recursive=True)
        zarr.consolidate_metadata(
            fs.get_mapper(dirname),  # type: ignore[arg-type]
        )
        # Invalidate any cached directory information.
        fs.invalidate_cache(dirname)


def add_zarr_array(
    dirname: str,
    variable: meta.Variable,
    template: str,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Add a variable to a Zarr dataset.

    Args:
        dirname: The name of the dataset.
        variable: The variable to add.
        template: The name of the template variable.
        fs: The file system that the dataset is stored on.
    """
    _LOGGER.debug('Adding variable %r to Zarr dataset %r', variable.name,
                  dirname)
    shape = zarr.open(fs.get_mapper(join_path(dirname, template))).shape
    store = fs.get_mapper(join_path(dirname, variable.name))
    zarr.create(
        shape,
        chunks=True,
        dtype=variable.dtype,
        compressor=variable.compressor,  # type: ignore[arg-type]
        fill_value=variable.fill_value,  # type: ignore[arg-type]
        store=store,
        filters=variable.filters)
    write_zattrs(dirname, variable, fs)
    zarr.consolidate_metadata(fs.get_mapper(dirname))  # type: ignore[arg-type]



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Partitioning scheme.
====================
"""
from __future__ import annotations

from typing import Any, ClassVar, Iterator, Sequence, Tuple
import abc
import collections
import re

import dask.array
import dask.array.core
import dask.array.creation
import dask.array.reductions
import dask.array.wrap
import fsspec
import numpy

from .. import dataset
from ..type_hints import NDArray

#: Object that represents a partitioning scheme
Partition = Tuple[Tuple[Tuple[str, Any], ...], slice]

#: Allowed data types for partitioning schemes
DATA_TYPES = ('int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32',
              'uint64')


def _logical_or_reduce(
    arr: dask.array.core.Array,
    axis: int | tuple[int, ...] | None = None,
) -> dask.array.core.Array:
    """Implementation of `numpy.logical_or` reduction with dask.

    Args:
        arr: Array to reduce.
        axis: Axis to reduce. If this is None, a reduction is performed over
            all the axes. If this is a tuple of ints, a reduction is performed
            on multiple axes, instead of a single axis or all the axes as
            before.

    Returns:
        Reduced array.
    """
    axis = axis or 0

    #: pylint: disable=unused-argument
    # The function signature is required by the `dask.array.reduce` function.
    def chunk(block, axis, keepdims):
        return block

    def aggregate(block, axis, keepdims):
        return numpy.logical_or.reduce(block, axis=axis)

    #: pylint: enable=unused-argument

    return dask.array.reductions.reduction(
        arr[1:] != arr[:-1],  # type: ignore
        chunk=chunk,
        aggregate=aggregate,
        axis=axis,
        keepdims=False,
        dtype=numpy.bool_)


def unique(arr: dask.array.core.Array) -> tuple[NDArray, NDArray]:
    """Return unique elements and their indices.

    Args:
        arr: Array of elements.

    Returns:
        Tuple of unique elements and their indices.
    """
    size = arr.shape[0]
    chunks = arr.chunks[0]
    #: pylint: disable=not-callable
    mask = dask.array.wrap.empty((size, ), dtype=numpy.bool_, chunks=chunks)
    #: pylint: enable=not-callable
    mask[0] = True
    mask[1:] = (_logical_or_reduce(arr, axis=1)
                if arr.ndim > 1 else arr[1:] != arr[:-1])
    dtype = numpy.uint32 if size < 2**32 else numpy.uint64
    indices = dask.array.creation.arange(size, dtype=dtype, chunks=chunks)
    mask = mask.persist()
    return arr[mask].compute(), indices[mask].compute()


def difference(arr: NDArray) -> NDArray:
    """Calculate the difference between each element in the array and the
    previous element.

    Args:
        arr: Array to calculate the difference for.

    Returns:
        Array of differences
    """
    return arr[1:] - arr[:-1]  # type: ignore


def concatenate_item(arr: NDArray, item: Any) -> NDArray:
    """Concatenate an array with a given item.

    Args:
        arr: Array to concatenate.
        item: Item to concatenate.

    Returns:
        Concatenated array.
    """
    return numpy.concatenate([arr, numpy.array([item], dtype=arr.dtype)])


def list_partitions(
    fs: fsspec.AbstractFileSystem,
    path: str,
    depth: int,
    root: bool = True,
) -> Iterator[str]:
    """The number of variables used for partitioning.

    The function will go down the tree and return all the files present when the
    requested depth is reached.

    Args:
        fs: file system object
        path: path to the directory
        depth: maximum depth of the directory tree.
        root: if True, the path is the root of the tree.

    Returns:
        Iterator of (path, directories, files).
    """
    if depth == -1:
        return StopIteration()

    if depth == 0:
        yield from sorted(fs.ls(path, detail=False))
        return StopIteration()

    if root:
        folders = map(
            lambda info: info['name'].rstrip('/'),
            filter(
                lambda info: info['type'] == 'directory' and not info['name'].
                split(fs.sep)[-1].startswith('.'),
                fs.ls(path, detail=True),
            ),
        )

        for pathname in sorted(folders):
            yield from list_partitions(fs,
                                       pathname,
                                       depth=depth - 1,
                                       root=False)
        return StopIteration()

    for item in sorted(fs.ls(path, detail=False)):
        yield from list_partitions(fs, item, depth=depth - 1, root=False)
    return StopIteration()


class Partitioning(abc.ABC):
    """Partitioning scheme.

    Args:
        variables:  List of variables to be used for partitioning
        dtype: The list of data types allowing to store the values of variables
            in a binary representation without loss of information.
            Defaults to int64.
    """
    __slots__ = ('_dtype', '_pattern', 'variables')

    #: The ID of the partitioning scheme
    ID: ClassVar[str | None] = None

    def __init__(self,
                 variables: Sequence[str],
                 dtype: Sequence[str] | None = None) -> None:
        if isinstance(dtype, str):
            raise TypeError('dtype must be a sequence of strings')
        if len(variables) == 0:
            raise ValueError('variables must not be empty')
        #: Variables to be used for the partitioning.
        self.variables = tuple(variables)
        #: Data type used to store variable values in a binary representation
        #: without data loss.
        self._dtype = dtype or ('int64', ) * len(self.variables)
        #: The regular expression that matches the partitioning scheme.
        self._pattern = self._regex().search

        if len(set(self._dtype) - set(DATA_TYPES)) != 0:
            raise ValueError(
                f"Data type must be one of {', '.join(DATA_TYPES)}.")

    def __len__(self) -> int:
        """Return the number of partitions."""
        return len(self._dtype)

    def dtype(self) -> tuple[tuple[str, str], ...]:
        """Return the data type of the partitioning scheme."""
        return tuple(zip(self._keys(), self._dtype))

    def _keys(self) -> Sequence[str]:
        """Return the different keys of a partition."""
        return self.variables

    def _regex(self) -> re.Pattern:
        """Return a regular expression that matches the partitioning scheme."""
        return re.compile('.'.join(f'({item})=(.*)' for item in self._keys()))

    @abc.abstractmethod
    def _split(
        self,
        variables: dict[str, dask.array.core.Array],
    ) -> Iterator[Partition]:
        """Split the variables constituting the partitioning into partitioning
        schemes.

        Args:
            variables: The variables to be split constituting the
                partitioning scheme.

        Returns:
            A sequence of tuples that contains the partitioning
                scheme and the associated indexer to divide the dataset on each
                partition found..
        """

    @staticmethod
    def _partition(selection: tuple[tuple[str, Any], ...]) -> tuple[str, ...]:
        """Format the partitioning scheme."""
        return tuple(f'{k}={v}' for k, v in selection)

    def index_dataset(
        self,
        ds: dataset.Dataset,
    ) -> Iterator[Partition]:
        """Yield the indexing scheme for the given dataset.

        Args:
            ds: The dataset to be indexed.

        Yields:
            The indexing scheme for the partitioning scheme.

        Raises:
            ValueError: if one of the variables needs for the partitioning
                is not monotonic.
        """
        variables = collections.OrderedDict(
            (name, ds.variables[name].array) for name in self.variables)
        # If the dask array is too chunked, the calculation is excessively
        # long.
        return self._split(
            {name: arr.rechunk().persist()
             for name, arr in variables.items()})

    def split_dataset(
        self,
        ds: dataset.Dataset,
        axis: str,
    ) -> Iterator[tuple[tuple[str, ...], dict[str, slice]]]:
        """Split the dataset into partitions.

        Args:
            ds:  The dataset to be split.
            axis: The axis to be used for the splitting.

        Yields:
            The partitioning scheme and the indexer to divide the dataset on
            each partition found.

        Raises:
            ValueError: if one of the variables needs for the partitioning
                is not a one-dimensional array.
        """
        for item in self.variables:
            if len(ds.variables[item].shape) != 1:
                raise ValueError(f'f{item!r} must be a one-dimensional array')
        return ((self._partition(selection), {
            axis: indexer
        }) for selection, indexer in self.index_dataset(ds))

    def get_config(self) -> dict[str, Any]:
        """Return the configuration of the partitioning scheme.

        Returns:
            The configuration of the partitioning scheme.
        """
        config = {'id': self.ID}
        slots = (getattr(_class, '__slots__', ())
                 for _class in reversed(self.__class__.__mro__))
        config.update((attr, getattr(self, attr)) for _class in slots
                      for attr in _class if not attr.startswith('_'))
        return config

    @classmethod
    def from_config(cls, config) -> Partitioning:
        """Create a partitioning scheme from a configuration.

        Args:
            config: The configuration of the partitioning scheme.

        Returns:
            The partitioning scheme.
        """
        return cls(**config)

    def parse(self, partition: str) -> tuple[tuple[str, int], ...]:
        """Parse a partitioning scheme.

        Args:
            partition: The partitioning scheme to be parsed.

        Returns:
            The parsed partitioning scheme.
        """
        match = self._pattern(partition)
        if match is None:
            raise ValueError(
                f'Partition is not driven by this instance: {partition}')
        groups = match.groups()
        return tuple((groups[ix], int(groups[ix + 1]))
                     for ix in range(0, len(groups), 2))

    @abc.abstractmethod
    def encode(
        self,
        partition: tuple[tuple[str, int], ...],
    ) -> tuple[Any, ...]:
        """Encode a partitioning scheme to the handled values.

        Args:
            partition: The partitioning scheme to be encoded.

        Returns:
            The encoded partitioning scheme.
        """

    @abc.abstractmethod
    def decode(self, values: tuple[Any, ...]) -> tuple[tuple[str, int], ...]:
        """Decode a partitioning scheme.

        Args:
            values: The encoded partitioning scheme.

        Returns:
            The decoded partitioning scheme.
        """

    @staticmethod
    def join(partition_scheme: tuple[tuple[str, int], ...], sep: str) -> str:
        """Join a partitioning scheme.

        Args:
            partition_scheme: The partitioning scheme to be joined.
            sep: The separator to be used.

        Returns:
            The joined partitioning scheme.
        """
        return sep.join(f'{k}={v}' for k, v in partition_scheme)

    def list_partitions(
        self,
        fs: fsspec.AbstractFileSystem,
        path: str,
    ) -> Iterator[str]:
        """List the partitions.

        Args:
            fs: The filesystem to be used.
            path: The path to the directory containing the partitions.

        Yields:
            The partitions.
        """
        return list_partitions(fs, path, depth=len(self) - 1)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Dataset variable.
=================
"""
from __future__ import annotations

from typing import Any, Iterable, Iterator, Mapping, MutableMapping, Sequence
import collections
import uuid

import dask.array.core
import dask.array.creation
import dask.array.ma
import dask.array.routines
import dask.array.wrap
import dask.base
import dask.threaded
import numcodecs.abc
import numpy
import xarray
import zarr

from . import mathematics, meta
from .meta import Attribute
from .type_hints import ArrayLike, NDArray, NDMaskedArray

#: The dask array getter used to access the data.
GETTER = dask.array.core.getter


def _dimensions_repr(dimensions: dict[str, int]) -> str:
    """Get the string representation of the dimensions.

    Args:
        dimensions: The dimensions.

    Returns:
        The string representation of the dimensions.
    """
    return str(tuple(f'{name}: {value}' for name, value in dimensions.items()))


def _calculate_column_width(items: Iterable) -> int:
    """Calculate the maximum width of a column.

    Args:
        items: An iterable of items.

    Returns:
        The maximum width of a column.
    """
    max_name = max(len(str(name)) for name in items)
    return max(max_name, 7)


def _maybe_truncate(obj: Any, max_size: int) -> str:
    """Truncate the string representation of an object to the given length.

    Args:
        obj: An object.
        max_size: The maximum length of the string representation.

    Returns:
        The string representation of the object.
    """
    result = str(obj)
    if len(result) > max_size:
        return result[:max_size - 3] + '...'
    return result


def _pretty_print(obj: Any, num_characters: int = 120) -> str:
    """Pretty print the object.

    Args:
        obj: An object.
        num_characters: The maximum number of characters per line.

    Returns:
        The pretty printed string representation of the object.
    """
    result = _maybe_truncate(obj, num_characters)
    return result + ' ' * max(num_characters - len(result), 0)


def _attributes_repr(attrs: Sequence[Attribute]) -> Iterator[str]:
    """Get the string representation of the attributes.

    Args:
        attrs: The attributes.

    Returns:
        The string representation of the attributes.
    """
    width = _calculate_column_width(item.name for item in attrs)
    for attr in attrs:
        name_str = f'    {attr.name:<{width}s}'
        yield _pretty_print(f'{name_str}: {attr.value!r}')


def _variable_repr(var: Variable) -> str:
    """Get the string representation of a variable.

    Args:
        var: A variable.

    Returns:
        The string representation of the variable.
    """
    # Dimensions
    dims_str = _dimensions_repr(dict(zip(var.dimensions, var.shape)))
    lines = [
        f'<{var.__module__}.{var.__class__.__name__} {dims_str}>',
        f'{var.array!r}'
    ]
    # Attributes
    if len(var.attrs):
        lines.append('  Attributes:')
        lines += _attributes_repr(var.attrs)
    # Filters
    if var.filters:
        lines.append('  Filters:')
        lines += [f'    {item!r}' for item in var.filters]
    # Compressor
    if var.compressor:
        lines.append('  Compressor:')
        lines += [f'    {var.compressor!r}']
    return '\n'.join(lines)


def _not_equal(first: Any, second: Any) -> bool:
    """Check if two objects are not equal.

    Args:
        first: The first object.
        second: The second object.

    Returns:
        True if the two objects are different, False otherwise.
    """

    def _is_not_a_number(number: Any) -> bool:
        """Check if a number is NaN or NaT."""
        # pylint: disable=comparison-with-itself
        return not number == number and number != number
        # pylint: enable=comparison-with-itself

    #: pylint: disable=unidiomatic-typecheck
    if type(first) != type(second):
        return True
    #: pylint: enable=unidiomatic-typecheck
    if _is_not_a_number(first) and _is_not_a_number(second):
        return False
    if first is None and second is None:
        return False
    if first == second:
        return False
    return True


def _asarray(
    arr: ArrayLike[Any],
    fill_value: Any | None = None,
) -> tuple[dask.array.core.Array, Any]:
    """Convert an array-like object to a dask array.

    Args:
        arr: An array-like object.
        fill_value: The fill value.

    Returns:
        If the data provided is a masked array, the functions return the array
        with masked data replaced by its fill value and the fill value of the
        offered masked array. Otherwise, the provided array and fill value.
    """
    result = dask.array.core.asarray(arr)  # type: dask.array.core.Array
    _meta = result._meta  # pylint: disable=protected-access
    if isinstance(_meta, numpy.ma.MaskedArray):
        if fill_value is not None and _not_equal(fill_value, _meta.fill_value):
            raise ValueError(
                f'The fill value {fill_value!r} does not match the fill value '
                f'{_meta.fill_value!r} of the array.')
        return dask.array.ma.filled(result, _meta.fill_value), _meta.fill_value
    return result, fill_value


def _new_variable(
    name: str,
    data: dask.array.core.Array,
    dimensions: Sequence[str],
    attrs: Sequence[Attribute],
    compressor: numcodecs.abc.Codec | None,
    fill_value: Any | None,
    filters: Sequence[numcodecs.abc.Codec] | None,
) -> Variable:
    """Create a new variable.

    Args:
        name: Name of the variable
        data: Variable data
        dimensions: Variable dimensions
        attrs: Variable attributes
        compressor: Compression codec
        fill_value: Value to use for uninitialized values
        filters: Filters to apply before writing data to disk
    """
    self = Variable.__new__(Variable)
    self.array = data
    self.attrs = attrs
    self.compressor = compressor
    self.dimensions = dimensions
    self.fill_value = fill_value
    self.filters = filters
    self.name = name
    return self


def _blockdims_from_blockshape(
        shape: tuple[int, ...],
        chunks: tuple[int, ...]) -> tuple[tuple[int, ...], ...]:
    """Convert a blockshape to a blockdims tuple."""
    return tuple(((chunk_item, ) * (shape_item // chunk_item) +
                  ((shape_item % chunk_item, ) if shape_item %
                   chunk_item else ()) if shape_item else (0, ))
                 for shape_item, chunk_item in zip(shape, chunks))


def _from_zarr_array(
    array: zarr.Array,
    shape: Sequence[int],
    chunks: Sequence[int],
    name: str,
    lock: bool = False,
    asarray=True,
    inline_array=True,
) -> dask.array.core.Array:
    """Create a dask array from a zarr array.

    Args:
        array: A zarr array.

    Returns:
        The dask array.
    """
    normalized_chunks = sum(
        (_blockdims_from_blockshape(
            (shape_item, ),
            (chunk_item, )) if not isinstance(chunk_item, (tuple, list)) else
         (chunk_item, ) for shape_item, chunk_item in zip(shape, chunks)),
        (),
    )
    dsk = dask.array.core.graph_from_arraylike(
        array,
        normalized_chunks,
        shape,
        name,
        getitem=GETTER,
        lock=lock,
        asarray=asarray,
        dtype=array.dtype,
        inline_array=inline_array,
    )
    return dask.array.core.Array(dsk,
                                 name,
                                 normalized_chunks,
                                 meta=array,
                                 dtype=array.dtype)


class Variable:
    """Variables hold multi-dimensional arrays of data.

    Args:
        name: Name of the variable
        data: Variable data
        dimensions: Variable dimensions
        attrs: Variable attributes
        compressor: Compression codec
        fill_value: Value to use for uninitialized values
        filters: Filters to apply before writing data to disk
    """
    __slots__ = ('array', 'attrs', 'compressor', 'dimensions', 'fill_value',
                 'filters', 'name')

    def __init__(self,
                 name: str,
                 data: ArrayLike[Any],
                 dimensions: Sequence[str],
                 attrs: Sequence[Attribute] | None = None,
                 compressor: numcodecs.abc.Codec | None = None,
                 fill_value: Any | None = None,
                 filters: Sequence[numcodecs.abc.Codec] | None = None) -> None:
        array, fill_value = _asarray(data, fill_value)
        #: Variable name
        self.name = name
        #: Variable data as a dask array.
        self.array: dask.array.core.Array = array
        #: Variable dimensions
        self.dimensions = dimensions
        #: Variable attributes
        self.attrs: Sequence[Attribute] = attrs or tuple()
        #: Compressor used to compress the data during writing data to disk
        self.compressor = compressor
        #: Value to use for uninitialized values
        self.fill_value = fill_value
        #: Filters to apply before writing data to disk
        self.filters = filters

    @property
    def dtype(self) -> numpy.dtype:
        """Return the data type of the variable."""
        return self.array.dtype

    @property
    def ndim(self) -> int:
        """Return the number of dimensions of the variable."""
        return len(self.dimensions)

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the variable."""
        return self.array.shape

    @property
    def size(self: Any) -> int:
        """Return the size of the variable."""
        return mathematics.prod(self.shape)

    @property
    def nbytes(self):
        """Return the number of bytes used by the variable."""
        return self.size * self.dtype.itemsize

    def metadata(self) -> meta.Variable:
        """Get the variable metadata.

        Returns:
            Variable metadata
        """
        return meta.Variable(self.name, self.dtype, self.dimensions,
                             self.attrs, self.compressor, self.fill_value,
                             self.filters)

    def persist(self, **kwargs) -> Variable:
        """Persist the variable data into memory.

        Args:
            **kwargs: Keyword arguments passed to
                :meth:`dask.array.Array.persist`.

        Returns:
            The variable
        """
        self.array = dask.base.persist(self.data, **kwargs)  # type: ignore
        return self

    @property
    def data(self) -> dask.array.core.Array:
        """Return the underlying dask array where values equal to the fill
        value are masked. If no fill value is set, the returned array is the
        same as the underlying array.

        Returns:
            The dask array

        .. seealso::

            :meth:`Variable.array`
        """
        # If the fill value is None, or if the dask array already holds a
        # masked array, return the underlying array.
        # pylint: disable=protected-access
        # No other way to check if the dask array is a masked array.
        if (self.fill_value is None
                or isinstance(self.array._meta, numpy.ma.MaskedArray)):
            return self.array
        # pylint: enable=protected-access
        return dask.array.ma.masked_equal(self.array, self.fill_value)

    @data.setter
    def data(self, data: Any) -> None:
        """Defines the underlying dask array. If the data provided is a masked
        array, it's converted to an array, where the masked values are replaced
        by its fill value, and its fill value becomes the new fill value of
        this instance. Otherwise, the underlying array is defined as the new
        data and the fill value is set to None.

        Args:
            data: The new data to use

        Raises:
            ValueError: If the shape of the data does not match the shape of
                the stored data.
        """
        data, fill_value = _asarray(data, self.fill_value)
        if len(data.shape) != len(self.dimensions):
            raise ValueError('data shape does not match variable dimensions')
        self.array, self.fill_value = data, fill_value

    @property
    def values(self) -> NDArray | NDMaskedArray:
        """Return the variable data as a numpy array.

        .. note::

            If the variable has a fill value, the result is a masked array where
            masked values are equal to the fill value.

        Returns:
            The variable data
        """
        return self.compute()

    def compute(self, **kwargs) -> NDArray | NDMaskedArray:
        """Return the variable data as a numpy array.

        .. note::

            If the variable has a fill value, the result is a masked array where
            masked values are equal to the fill value.

        Args:
            **kwargs: Keyword arguments passed to
                :meth:`dask.array.Array.compute`.
        """
        (values, ) = dask.base.compute(self.array, traverse=False, **kwargs)
        return values if self.fill_value is None else numpy.ma.masked_equal(
            values, self.fill_value)

    def fill(self) -> Variable:
        """Fill the variable with the fill value. If the variable has no fill
        value, this method does nothing.

        Returns:
            The variable.
        """
        if self.fill_value is not None:
            self.array = dask.array.creation.full_like(self.array,
                                                       self.fill_value)
        return self

    def isel(self, key: tuple[slice, ...]) -> Variable:
        """Return a new variable with data selected along the given dimension
        indices.

        Args:
            key: Dimension indices to select

        Returns:
            The new variable
        """
        return _new_variable(self.name, self.array[key], self.dimensions,
                             self.attrs, self.compressor, self.fill_value,
                             self.filters)

    @classmethod
    def from_zarr(cls, array: zarr.Array, name: str, dimension: str,
                  **kwargs) -> Variable:
        """Create a new variable from a zarr array.

        Args:
            array: The zarr array
            name: Name of the variable
            dimension: Name of the attribute that defines the dimensions of the
                variable
            **kwargs: Keyword arguments passed to
                :func:`dask.array.from_array`

        Returns:
            The variable
        """
        attrs = tuple(
            Attribute(k, v) for k, v in array.attrs.items() if k != dimension)
        data = _from_zarr_array(
            array,
            array.shape,
            array.chunks,
            name=f'{name}-{uuid.uuid1()}',
            **kwargs,
        )
        return _new_variable(name, data, array.attrs[dimension], attrs,
                             array.compressor, array.fill_value, array.filters)

    def duplicate(self, data: Any) -> Variable:
        """Create a new variable from the properties of this instance and the
        data provided.

        Args:
            data: Variable data.

        Returns:
            New variable.

        Raises:
            ValueError: If the shape of the data does not match the shape of
                the stored data.
        """
        result = Variable(self.name, data, self.dimensions, self.attrs,
                          self.compressor, self.fill_value, self.filters)
        if len(result.shape) != len(self.dimensions):
            raise ValueError('data shape does not match variable dimensions')
        return result

    def rename(self, name: str) -> Variable:
        """Rename the variable.

        Args:
            name: New variable name.

        Returns:
            The variable.
        """
        return _new_variable(name, self.array, self.dimensions, self.attrs,
                             self.compressor, self.fill_value, self.filters)

    def dimension_index(self) -> Iterator[tuple[str, int]]:
        """Return an iterator over the variable dimensions and their index.

        Returns:
            An iterator over the variable dimensions
        """
        yield from ((item, ix) for ix, item in enumerate(self.dimensions))

    def concat(self, other: Variable | Sequence[Variable],
               dim: str) -> Variable:
        """Concatenate this variable with another variable or a list of
        variables along a dimension.

        Args:
            other: Variable or list of variables to concatenate with this
                variable.
            dim: Dimension to concatenate along.

        Returns:
            New variable.

        Raises:
            ValueError: if the variables provided is an empty sequence.
        """
        if not isinstance(other, Sequence):
            other = [other]
        if not other:
            raise ValueError('other must be a non-empty sequence')
        try:
            axis = self.dimensions.index(dim)
            return _new_variable(
                self.name,
                dask.array.core.concatenate(
                    (self.array, *(item.array for item in other)), axis=axis),
                self.dimensions,
                self.attrs,
                self.compressor,
                self.fill_value,
                self.filters,
            )
        except ValueError:
            # If the concatenation dimension is not within the dimensions of the
            # variable, then the original variable is returned (i.e.
            # concatenation is not necessary).
            return _new_variable(self.name, self.array, self.dimensions,
                                 self.attrs, self.compressor, self.fill_value,
                                 self.filters)

    def rechunk(self, **kwargs) -> Variable:
        """Rechunk the variable.

        Args:
            **kwargs: Keyword arguments passed to
                :func:`dask.array.rechunk`

        Returns:
            The variable.
        """
        return _new_variable(self.name, self.array.rechunk(**kwargs),
                             self.dimensions, self.attrs, self.compressor,
                             self.fill_value, self.filters)

    def to_xarray(self) -> xarray.Variable:
        """Convert the variable to an xarray.Variable.

        Returns:
            Variable as an xarray.Variable
        """
        encoding = {}
        if self.filters:
            encoding['filters'] = self.filters
        if self.compressor:
            encoding['compressor'] = self.compressor
        data = self.array
        if self.dtype.kind == 'M':
            # xarray need a datetime64[ns] dtype
            data = data.astype('datetime64[ns]')
            encoding['dtype'] = 'int64'
        elif self.dtype.kind == 'm':
            encoding['dtype'] = 'int64'
        attrs = collections.OrderedDict(
            (item.name, item.value) for item in self.attrs)
        if self.fill_value is not None:
            attrs['_FillValue'] = self.fill_value
        return xarray.Variable(self.dimensions, data, attrs, encoding)

    def __str__(self) -> str:
        return _variable_repr(self)

    def __repr__(self) -> str:
        return _variable_repr(self)

    def __hash__(self) -> int:
        return hash(self.name)

    def __getitem__(self, key: Any) -> Any:
        return self.data[key]

    def __array__(self):
        return self.values

    def to_dask_array(self):
        """Return the underlying dask array.

        Returns:
            The underlying dask array

        .. seealso::

            :func:`dask.array.asarray`
        """
        return self.array

    def __dask_graph__(self) -> Mapping | None:
        """Return the dask Graph."""
        return self.array.__dask_graph__()

    def __dask_keys__(self) -> list:
        """Return the output keys for the Dask graph."""
        return self.array.__dask_keys__()

    def __dask_layers__(self) -> tuple:
        """Return the layers for the Dask graph."""
        return self.array.__dask_layers__()

    def __dask_tokenize__(self):
        """Return the token for the Dask graph."""
        return dask.base.normalize_token(
            (type(self), self.name, self.array, self.dimensions, self.attrs,
             self.fill_value))

    @staticmethod
    def __dask_optimize__(dsk: MutableMapping, keys: list,
                          **kwargs) -> MutableMapping:
        """Returns whether the Dask graph can be optimized.

        .. seealso::
            :meth:`dask.array.Array.__dask_optimize__`
        """
        return dask.array.core.Array.__dask_optimize__(dsk, keys, **kwargs)

    #: The default scheduler get to use for this object.
    __dask_scheduler__ = staticmethod(dask.threaded.get)

    def _dask_finalize(self, results, array_func, *args, **kwargs):
        array = array_func(results, *args, **kwargs)
        if not isinstance(array, dask.array.core.Array):
            array = dask.array.core.from_array(array)
        return _new_variable(self.name, array, self.dimensions, self.attrs,
                             self.compressor, self.fill_value, self.filters)

    def __dask_postcompute__(self) -> tuple:
        """Return the finalizer and extra arguments to convert the computed
        results into their in-memory representation."""
        array_func, array_args = self.array.__dask_postcompute__()
        return self._dask_finalize, (array_func, ) + array_args

    def __dask_postpersist__(self) -> tuple:
        """Return the rebuilder and extra arguments to rebuild an equivalent
        Dask collection from a persisted or rebuilt graph."""
        array_func, array_args = self.array.__dask_postpersist__()
        return self._dask_finalize, (array_func, ) + array_args



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Registers the partitioning codecs.
==================================
"""
from __future__ import annotations

from . import abc

#: A registry of all available partitioning codecs.
CODEC_REGISTRY: dict[str, abc.Partitioning] = {}


def get_codecs(config: dict) -> abc.Partitioning:
    """Get the partitioning scheme for the given configuration.

    Args:
        config: A dictionary of the partitioning configuration parameters.

    Returns:
        The partitioning scheme.

    Raises:
        ValueError: If the requested codec is not defined.
    """
    codec_id = config.pop('id', None)
    if codec_id is None:
        raise ValueError(f'codec not available: {codec_id!r}')
    cls = CODEC_REGISTRY.get(codec_id, None)
    if cls is None:
        raise ValueError(f'codec not available: {codec_id!r}')
    return cls.from_config(config)


def register_codec(cls, codec_id=None) -> None:
    """Register a partitioning scheme.

    Args:
        cls: The partitioning scheme class.
        codec_id: The partitioning scheme identifier.

    Raises:
        ValueError: If the codec identifier is already registered.
    """
    if codec_id is None:
        codec_id = cls.ID
    if codec_id in CODEC_REGISTRY:
        raise ValueError(f'codec already registered: {codec_id!r}')
    CODEC_REGISTRY[codec_id] = cls



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Convenience functions
=====================
"""
from .collection import create_collection, open_collection
from .view import create_view, open_view

__all__ = ['create_collection', 'open_collection', 'create_view', 'open_view']



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing datasets
================
"""
import pickle

import numpy
import pytest
import xarray
import zarr

from .. import dataset, meta
# pylint: disable=unused-import # Need to import for fixtures
from .cluster import dask_client, dask_cluster
# pylint enable=unused-import
from .test_variable import create_test_variable


def create_test_dataset():
    """Create a test dataset."""
    return dataset.Dataset(attrs=(dataset.Attribute(name='attr', value=1), ),
                           variables=(create_test_variable(),
                                      create_test_variable('var2')))


def test_dataset(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test dataset creation."""
    ds = create_test_dataset()
    assert ds.dimensions == dict(x=5, y=2)
    assert ds.attrs == (dataset.Attribute(name='attr', value=1), )
    assert isinstance(str(ds), str)
    assert isinstance(repr(ds), str)
    assert ds.nbytes == 5 * 2 * 8 * 2

    var1 = create_test_variable()
    assert numpy.all(ds.variables['var1'].values == var1.values)
    assert ds.variables['var1'].metadata() == var1.metadata()
    assert numpy.all(var1.values == ds['var1'].values)
    assert var1.metadata() == ds['var1'].metadata()
    with pytest.raises(KeyError):
        var1 = ds['varX']

    var2 = create_test_variable('var2')
    assert numpy.all(ds.variables['var2'].values == var2.values)
    assert id(ds['var2']) == id(ds.variables['var2'])
    assert ds.variables['var2'].metadata() == var2.metadata()
    assert isinstance(ds.metadata(), meta.Dataset)

    other = ds.compute()
    assert isinstance(other, dataset.Dataset)

    ds_dict = ds.to_dict()
    assert isinstance(ds_dict, dict)
    assert isinstance(ds_dict['var1'], numpy.ndarray)
    assert isinstance(ds_dict['var2'], numpy.ndarray)

    ds_dict = ds.to_dict(variables=['var1'])
    assert isinstance(ds_dict, dict)
    assert isinstance(ds_dict['var1'], numpy.ndarray)
    assert 'var2' not in ds_dict


def test_dataset_dimensions_conflict(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test dataset creation with dimensions conflict."""
    with pytest.raises(ValueError):
        dataset.Dataset([
            dataset.Variable(name='var1',
                             data=numpy.arange(10,
                                               dtype='int64').reshape(5, 2),
                             dimensions=('x', 'y'),
                             attrs=(dataset.Attribute(name='attr', value=1), ),
                             compressor=zarr.Blosc(cname='zstd', clevel=1),
                             fill_value=0,
                             filters=(zarr.Delta('int64', 'int32'),
                                      zarr.Delta('int32', 'int32'))),
            dataset.Variable(name='var2',
                             data=numpy.arange(20, dtype='int64'),
                             dimensions=('x'),
                             attrs=(dataset.Attribute(name='attr', value=1), ),
                             compressor=zarr.Blosc(cname='zstd', clevel=1),
                             fill_value=0,
                             filters=(zarr.Delta('int64', 'int32'),
                                      zarr.Delta('int32', 'int32'))),
        ])


def test_xarray(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test xarray creation."""
    ds = create_test_dataset()
    xr1 = ds.to_xarray()
    assert isinstance(xr1, xarray.Dataset)

    xr2 = dataset.Dataset.from_xarray(xr1).to_xarray()
    assert xr1 == xr2


def test_dataset_isel(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test dataset selection."""
    ds = create_test_dataset()
    selected_ds = ds.isel(slices=dict(x=slice(0, 2)))
    assert selected_ds.dimensions == dict(x=2, y=2)
    assert selected_ds.attrs == (dataset.Attribute(name='attr', value=1), )
    assert numpy.all(
        selected_ds.variables['var1'].values == numpy.arange(4).reshape(2, 2))
    selected_ds = ds.isel(slices=dict(y=slice(0, 1)))
    assert selected_ds.dimensions == dict(x=5, y=1)
    assert numpy.all(selected_ds.variables['var1'].values == numpy.arange(
        0, 10, 2).reshape(5, 1))

    # Cannot slice on something which is not a dimension
    with pytest.raises(ValueError, match='invalid dimension'):
        ds.isel(slices=dict(z=slice(0, 1)))

    with pytest.raises(ValueError, match='invalid dimension'):
        ds.isel(slices=dict(var1=slice(0, 1)))


def test_dataset_delete(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test dataset deletion."""
    ds = create_test_dataset()

    other = ds.delete([1], 'y')
    assert numpy.all(
        other.variables['var1'].values == numpy.arange(0, 10, 2).reshape(5, 1))
    assert numpy.all(
        other.variables['var2'].values == numpy.arange(0, 10, 2).reshape(5, 1))

    other = ds.delete([0], 'x')
    assert numpy.all(other.variables['var1'].values == numpy.arange(
        10).reshape(5, 2)[1:, :])


def test_dataset_concat(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test concatenation of datasets."""
    ds1 = create_test_dataset()
    ds2 = create_test_dataset()
    ds3 = ds1.concat(ds2, 'y')

    matrix = numpy.arange(10).reshape(5, 2)
    assert numpy.all(ds3.variables['var1'].values == numpy.concatenate(
        (matrix, matrix), axis=1))

    ds3 = ds1.concat(ds2, 'x')
    assert numpy.all(ds3.variables['var1'].values == numpy.concatenate(
        (matrix, matrix), axis=0))

    with pytest.raises(ValueError):
        ds1.concat([], 'z')


def test_variable_pickle(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test pickling of variables."""
    variable = create_test_variable()
    other = pickle.loads(pickle.dumps(variable))
    assert numpy.all(variable.values == other.values)
    assert variable.attrs == other.attrs
    assert variable.compressor == other.compressor
    assert variable.dimensions == other.dimensions
    assert variable.dtype == other.dtype
    assert variable.fill_value == other.fill_value
    assert variable.filters == other.filters
    assert variable.name == other.name


def test_dataset_pickle(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test pickling of datasets."""
    ds = create_test_dataset()
    other = pickle.loads(pickle.dumps(ds))
    assert ds.attrs == other.attrs
    assert list(ds.variables) == list(other.variables)
    for varname, variable in ds.variables.items():
        assert variable.attrs == other.variables[varname].attrs
        assert variable.compressor == other.variables[varname].compressor
        assert variable.dimensions == other.variables[varname].dimensions
        assert variable.dtype == other.variables[varname].dtype
        assert variable.fill_value == other.variables[varname].fill_value
        assert variable.filters == other.variables[varname].filters
        assert variable.name == other.variables[varname].name


def test_dataset_add_variable(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test for adding a variable."""
    ds = create_test_dataset()
    var = meta.Variable('var2', numpy.int64,
                        ('x', 'y'), (dataset.Attribute('attr', 1), ),
                        zarr.Blosc(), 255, (zarr.Delta('int64', 'int64'), ))
    ds.add_variable(var)

    assert ds.variables['var2'].attrs == (dataset.Attribute(name='attr',
                                                            value=1), )
    assert ds.variables['var2'].dtype == numpy.int64
    assert ds.variables['var2'].dimensions == ('x', 'y')
    assert ds.variables['var2'].compressor == zarr.Blosc()
    assert ds.variables['var2'].filters == (zarr.Delta('int64', 'int64'), )
    assert ds.variables['var2'].fill_value == 255
    assert ds.variables['var2'].name == 'var2'
    assert ds.variables['var2'].shape == (5, 2)
    assert numpy.ma.allequal(
        ds.variables['var2'].values,
        numpy.ma.masked_equal(numpy.full((5, 2), 255, 'int64'), 255))

    other = ds.select_vars(['var1'])
    assert list(other.variables) == ['var1']

    data = numpy.ones((5, 2), 'int64')
    ds.drops_vars('var2')
    assert 'var2' not in ds.variables

    var = meta.Variable('var2', numpy.int64,
                        ('x', 'y'), (dataset.Attribute('attr', 1), ),
                        zarr.Blosc(), 255, (zarr.Delta('int64', 'int64'), ))
    ds.add_variable(var, data)
    assert numpy.ma.allequal(
        ds.variables['var2'].values,
        numpy.ma.masked_equal(numpy.full((5, 2), 1, 'int64'), 255))

    data = numpy.ones((10, 2), 'int64')
    with pytest.raises(ValueError, match='Conflicting sizes'):
        ds.add_variable(var, data)

    var = meta.Variable('var2', numpy.int64,
                        ('xx', 'y'), (dataset.Attribute('attr', 1), ),
                        zarr.Blosc(), 255, (zarr.Delta('int64', 'int64'), ))
    with pytest.raises(ValueError, match='has dimension'):
        ds.add_variable(var)


def test_empty_dataset(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test empty dataset."""
    ds = dataset.Dataset([], [])
    assert ds.attrs == ()
    assert list(ds.variables) == []
    assert (str(ds)) == """<zcollection.dataset.Dataset>
  Dimensions: ()
Data variables:
    <empty>"""


def test_dataset_rename(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test renaming of datasets."""
    ds = create_test_dataset()
    ds.rename(dict(var1='var3', var2='var4'))
    assert ds.variables['var3'].name == 'var3'
    assert ds.variables['var4'].name == 'var4'
    assert 'var1' not in ds.variables
    assert 'var2' not in ds.variables

    with pytest.raises(ValueError):
        ds.rename(dict(var3='var4'))


def test_dataset_persist(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test persisting of datasets."""
    ds1 = create_test_dataset()
    ds1.persist()
    ds1.persist()
    assert ds1.variables['var1'].values is not None
    assert ds1.variables['var2'].values is not None

    ds2 = create_test_dataset()
    ds2.persist(compress=True)
    ds2.persist(compress=True)
    assert ds2.variables['var1'].values is not None
    assert ds2.variables['var2'].values is not None

    assert numpy.all(
        ds1.variables['var1'].values == ds2.variables['var1'].values)
    assert numpy.all(
        ds1.variables['var2'].values == ds2.variables['var2'].values)


def test_dataset_rechunk(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test rechunking of datasets."""
    ds1 = create_test_dataset()
    ds2 = create_test_dataset()
    ds2 = ds2.rechunk()

    assert numpy.all(
        ds1.variables['var1'].values == ds2.variables['var1'].values)
    assert numpy.all(
        ds1.variables['var2'].values == ds2.variables['var2'].values)

    ds2 = ds1.rechunk().persist(compress=True)
    assert numpy.all(
        ds1.variables['var1'].values == ds2.variables['var1'].values)
    assert numpy.all(
        ds1.variables['var2'].values == ds2.variables['var2'].values)


def test_dataset_merge(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    template = dataset.Dataset([
        dataset.Variable('var1', numpy.empty(10), ('x', ), ()),
        dataset.Variable('var2', numpy.empty(10), ('x', ), ()),
        dataset.Variable('var3', numpy.empty((10, 10)), ('x', 'y'), ()),
        dataset.Variable('var4', numpy.empty((10, 10)), ('x', 'y'), ()),
    ], [dataset.Attribute('attr1', 1),
        dataset.Attribute('attr2', 2)])
    ds2 = dataset.Dataset([
        dataset.Variable('var5', numpy.empty(10), ('x', ), ()),
        dataset.Variable('var6', numpy.empty(10), ('x', ), ()),
        dataset.Variable('var7', numpy.empty((10, 10)), ('x', 'y'), ()),
        dataset.Variable('var8', numpy.empty((10, 10)), ('x', 'y'), ()),
    ], [dataset.Attribute('attr3', 3),
        dataset.Attribute('attr4', 4)])
    ds1 = pickle.loads(pickle.dumps(template))
    ds1.merge(ds2)
    assert list(ds1.variables) == [
        'var1', 'var2', 'var3', 'var4', 'var5', 'var6', 'var7', 'var8'
    ]
    assert ds1.dimensions == {'x': 10, 'y': 10}
    assert ds1.variables['var1'].shape == (10, )
    assert ds1.variables['var2'].shape == (10, )
    assert ds1.variables['var3'].shape == (10, 10)
    assert ds1.variables['var4'].shape == (10, 10)
    assert ds1.variables['var5'].shape == (10, )
    assert ds1.variables['var6'].shape == (10, )
    assert ds1.variables['var7'].shape == (10, 10)
    assert ds1.variables['var8'].shape == (10, 10)
    assert ds1.attrs == (dataset.Attribute('attr1',
                                           1), dataset.Attribute('attr2', 2),
                         dataset.Attribute('attr3',
                                           3), dataset.Attribute('attr4', 4))

    ds2 = dataset.Dataset([
        dataset.Variable('var5', numpy.empty(10), ('a', ), ()),
        dataset.Variable('var6', numpy.empty((10, 10)), ('a', 'b'), ()),
    ], [dataset.Attribute('attr3', 3),
        dataset.Attribute('attr4', 4)])
    ds1 = pickle.loads(pickle.dumps(template))
    ds1.merge(ds2)
    assert list(
        ds1.variables) == ['var1', 'var2', 'var3', 'var4', 'var5', 'var6']
    assert ds1.dimensions == {'x': 10, 'y': 10, 'a': 10, 'b': 10}
    assert ds1.variables['var1'].shape == (10, )
    assert ds1.variables['var2'].shape == (10, )
    assert ds1.variables['var3'].shape == (10, 10)
    assert ds1.variables['var4'].shape == (10, 10)
    assert ds1.variables['var5'].shape == (10, )
    assert ds1.variables['var6'].shape == (10, 10)

    ds2 = dataset.Dataset([
        dataset.Variable('var5', numpy.empty(10), ('z', ), ()),
    ], [dataset.Attribute('attr3', 3),
        dataset.Attribute('attr4', 4)])
    ds1 = pickle.loads(pickle.dumps(template))
    ds1.merge(ds2)
    assert list(ds1.variables) == ['var1', 'var2', 'var3', 'var4', 'var5']
    assert ds1.dimensions == {'x': 10, 'y': 10, 'z': 10}
    assert ds1.variables['var1'].shape == (10, )
    assert ds1.variables['var2'].shape == (10, )
    assert ds1.variables['var3'].shape == (10, 10)
    assert ds1.variables['var4'].shape == (10, 10)
    assert ds1.variables['var5'].shape == (10, )

    ds2 = dataset.Dataset([
        dataset.Variable('var1', numpy.empty(10), ('x', ), ()),
    ], [dataset.Attribute('attr1', 1),
        dataset.Attribute('attr2', 2)])
    ds1 = pickle.loads(pickle.dumps(template))
    with pytest.raises(ValueError):
        ds1.merge(ds2)

    ds2 = dataset.Dataset([
        dataset.Variable('var5', numpy.empty(20), ('x', ), ()),
    ], [dataset.Attribute('attr3', 3),
        dataset.Attribute('attr4', 4)])
    ds1 = pickle.loads(pickle.dumps(template))
    with pytest.raises(ValueError):
        ds1.merge(ds2)


def test_dataset_select_variables_by_dims(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    ds = dataset.Dataset([
        dataset.Variable('var1', numpy.empty(10), ('x', ), ()),
        dataset.Variable('var2', numpy.empty(10), ('y', ), ()),
        dataset.Variable('var3', numpy.empty((10, 10)), ('x', 'y'), ()),
        dataset.Variable('var4', numpy.empty(20), ('a', ), ()),
        dataset.Variable('var5', numpy.empty(20), ('b', ), ()),
        dataset.Variable('var6', numpy.empty((20, 20)), ('a', 'b'), ()),
        dataset.Variable('var7', numpy.int64(1), (), ()),
    ], [
        dataset.Attribute('attr1', 1),
    ])
    selected = ds.select_variables_by_dims(('x', ))
    assert list(selected.variables) == ['var1', 'var3']
    assert selected.dimensions == {'x': 10, 'y': 10}

    selected = ds.select_variables_by_dims(('x', 'y'))
    assert list(selected.variables) == ['var1', 'var2', 'var3']
    assert selected.dimensions == {'x': 10, 'y': 10}

    selected = ds.select_variables_by_dims(('x', 'y', 'z'))
    assert list(selected.variables) == ['var1', 'var2', 'var3']
    assert selected.dimensions == {'x': 10, 'y': 10}

    selected = ds.select_variables_by_dims(('a', ))
    assert list(selected.variables) == ['var4', 'var6']
    assert selected.dimensions == {'a': 20, 'b': 20}

    selected = ds.select_variables_by_dims(('a', 'b'))
    assert list(selected.variables) == ['var4', 'var5', 'var6']
    assert selected.dimensions == {'a': 20, 'b': 20}

    selected = ds.select_variables_by_dims(('a', 'b', 'c'))
    assert list(selected.variables) == ['var4', 'var5', 'var6']
    assert selected.dimensions == {'a': 20, 'b': 20}

    selected = ds.select_variables_by_dims(('x', 'a'))
    assert list(selected.variables) == ['var1', 'var3', 'var4', 'var6']
    assert selected.dimensions == {'x': 10, 'y': 10, 'a': 20, 'b': 20}

    selected = ds.select_variables_by_dims(tuple())
    assert list(selected.variables) == ['var7']
    assert selected.dimensions == {}

    selected = ds.select_variables_by_dims(tuple(), predicate=False)
    assert list(selected.variables) == [
        'var1', 'var2', 'var3', 'var4', 'var5', 'var6'
    ]
    assert selected.dimensions == {'x': 10, 'y': 10, 'a': 20, 'b': 20}

    selected = ds.select_variables_by_dims(('x', ), predicate=False)
    assert list(selected.variables) == ['var2', 'var4', 'var5', 'var6', 'var7']



"""
Mathematical functions.
=======================
"""
from typing import Iterable
import functools
import operator


def prod(iterable: Iterable) -> int:
    """Get the product of an iterable.

    Args:
        iterable: An iterable.

    Returns:
        The product of the iterable.
    """
    return functools.reduce(operator.mul, iterable, 1)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Make test data.
===============
"""
import numpy
import xarray

START_DATE = numpy.datetime64('2000-01-01', 'us')
END_DATE = numpy.datetime64('2000-06-30', 'us')
DELTA = numpy.timedelta64(72, 'h')


def create_test_sequence(repeatability, number_of_measures, number_of_cycles):
    """Creation of a data set for testing purposes."""
    pass_number = numpy.hstack([
        numpy.tile(ix + 1, number_of_measures) for j in range(number_of_cycles)
        for ix in range(repeatability)
    ])
    cycle_number = numpy.hstack([
        numpy.tile(ix + 1, repeatability * number_of_measures)
        for ix in range(number_of_cycles)
    ])
    delta = numpy.timedelta64(24 // repeatability // 2, 'h')
    time = numpy.arange(START_DATE, START_DATE + len(cycle_number) * delta,
                        delta)
    observation = numpy.random.rand(cycle_number.size)  # type: ignore
    ds = xarray.Dataset(
        dict(time=xarray.DataArray(time, dims=('num_lines', )),
             cycle_number=xarray.DataArray(cycle_number, dims=('num_lines', )),
             pass_number=xarray.DataArray(pass_number, dims=('num_lines', )),
             observation=xarray.DataArray(observation, dims=('num_lines', ))))
    return ds



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test partitioning by date.
==========================
"""
from typing import Iterator
import pickle

import dask.array.core
import dask.local
import fsspec
import numpy
import pytest
import xarray

from .. import Date, get_codecs
from ... import dataset
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster

# pylint: disable=disable=unused-argument


def test_split_dataset(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test the split_dataset method."""
    start_date = numpy.datetime64('2000-01-06', 'us')
    delta = numpy.timedelta64(1, 'h')

    for end_date, indices, resolution in [
        (
            numpy.datetime64('2001-12-31', 'Y'),
            slice(0, 1),
            'Y',
        ),
        (
            numpy.datetime64('2000-12-31', 'M'),
            slice(0, 2),
            'M',
        ),
        (
            numpy.datetime64('2000-12-31', 'D'),
            slice(0, 3),
            'D',
        ),
        (
            numpy.datetime64('2000-01-31', 'h'),
            slice(0, 4),
            'h',
        ),
    ]:

        # Time delta between two partitions
        timedelta = numpy.timedelta64(1, resolution)

        # Temporal axis to split
        dates = numpy.arange(start_date, end_date, delta)

        # Measured data
        observation = numpy.random.rand(dates.size)  # type: ignore

        # Create the dataset to split
        ds = xarray.Dataset(
            dict(dates=xarray.DataArray(dates, dims=('num_lines', )),
                 observation=xarray.DataArray(observation,
                                              dims=('num_lines', ))))

        partitioning = Date(('dates', ), resolution)
        assert len(partitioning) == len(range(indices.start, indices.stop))

        # Date of the current partition
        date = numpy.datetime64(start_date, resolution)

        # Build the test dataset
        ds = dataset.Dataset.from_xarray(ds)

        iterator = partitioning.split_dataset(ds, 'num_lines')
        assert isinstance(iterator, Iterator)

        for partition, indexer in iterator:
            subset = ds.isel(indexer)

            # Cast the date to the a datetime object to extract the date
            item = date.astype('datetime64[us]').item()
            expected = (
                f'year={item.year}',
                f'month={item.month:02d}',
                f'day={item.day:02d}',
                f'hour={item.hour:02d}',
            )
            assert partition == expected[indices]

            folder = '/'.join(partition)
            fields = partitioning.parse(folder)
            parsed_date, = partitioning.encode(fields)
            assert parsed_date == numpy.datetime64(date).astype(
                f'datetime64[{resolution}]')

            expected_selection = dates[
                (dates >= parsed_date)  # type: ignore
                & (dates < parsed_date + timedelta)]  # type: ignore
            computed_selection = subset.variables['dates'].compute(
                scheduler=dask.local.get_sync)
            assert numpy.all(computed_selection == expected_selection)

            expected = (
                ('year', item.year),
                ('month', item.month),
                ('day', item.day),
                ('hour', item.hour),
            )
            assert fields == expected[indices]
            assert partitioning.join(fields, '/') == folder
            assert partitioning.join(partitioning.decode((parsed_date, )),
                                     '/') == folder

            date += timedelta


def test_construction():
    """Test the construction of the Date class."""
    partitioning = Date(('dates', ), 'D')
    assert partitioning.resolution == 'D'
    assert partitioning.variables == ('dates', )
    assert partitioning.dtype() == (('year', 'uint16'), ('month', 'uint8'),
                                    ('day', 'uint8'))
    assert len(partitioning) == 3
    assert partitioning.get_config() == {
        'id': 'Date',
        'resolution': 'D',
        'variables': ('dates', ),
    }

    with pytest.raises(ValueError):
        Date(('dates1', 'dates2'), 'D')

    with pytest.raises(ValueError):
        Date(('dates', ), 'W')


def test_config():
    """Test the configuration of the Date class."""
    partitioning = Date(('dates', ), 'D')
    assert partitioning.dtype() == (('year', 'uint16'), ('month', 'uint8'),
                                    ('day', 'uint8'))
    config = partitioning.get_config()
    partitioning = get_codecs(config)
    assert isinstance(partitioning, Date)


def test_pickle():
    """Test the pickling of the Date class."""
    partitioning = Date(('dates', ), 'D')
    other = pickle.loads(pickle.dumps(partitioning))
    assert isinstance(other, Date)
    assert other.resolution == 'D'
    assert other.variables == ('dates', )


def test_no_monotonic(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test that the Date partitioning raises an error if the temporal axis is
    not monotonic."""
    dates = numpy.arange(numpy.datetime64('2000-01-01', 'h'),
                         numpy.datetime64('2000-01-02', 'h'),
                         numpy.timedelta64(1, 'm'))
    numpy.random.shuffle(dates)
    partitioning = Date(('dates', ), 'h')
    # pylint: disable=protected-access
    with pytest.raises(ValueError):
        list(partitioning._split({'dates': dask.array.core.from_array(dates)}))
    # pylint: enable=protected-access


def test_values_must_be_datetime64(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test that the values must be datetime64."""
    dates = numpy.arange(numpy.datetime64('2000-01-01', 'h'),
                         numpy.datetime64('2000-01-02', 'h'),
                         numpy.timedelta64(1, 'm'))
    partitioning = Date(('dates', ), 'h')
    dates = dates.astype('int64')
    with pytest.raises(TypeError):
        # pylint: disable=protected-access
        list(partitioning._split({'dates': dask.array.core.from_array(dates)}))
    # pylint: enable=protected-access


def test_listing_partition():
    """Test the listing of the partitions."""
    fs = fsspec.filesystem('memory')
    variables = [
        'ancillary_surface_classification_flag',
        'correction_flag',
        'cross_track_angle',
        'cross_track_distance',
        'cycle_number',
        'dac',
        'depth_or_elevation',
        'distance_to_coast',
        'doppler_centroid',
        'dynamic_ice_flag',
        'geoid',
        'heading_to_coast',
        'height_cor_xover',
        'ice_conc',
        'internal_tide_hret',
        'internal_tide_sol2',
        'inv_bar_cor',
        'iono_cor_gim_ka',
        'latitude',
        'latitude_avg_ssh',
        'latitude_nadir',
        'load_tide_fes',
        'load_tide_got',
        'longitude',
        'longitude_avg_ssh',
        'longitude_nadir',
        'mean_dynamic_topography',
        'mean_dynamic_topography_uncert',
        'mean_sea_surface_cnescls',
        'mean_sea_surface_cnescls_uncert',
        'mean_sea_surface_dtu',
        'mean_sea_surface_dtu_uncert',
        'mean_wave_direction',
        'mean_wave_period_t02',
        'model_dry_tropo_cor',
        'model_wet_tropo_cor',
        'num_pt_avg',
        'obp_ref_surface',
        'ocean_tide_eq',
        'ocean_tide_fes',
        'ocean_tide_got',
        'ocean_tide_non_eq',
        'orbit_alt_rate',
        'orbit_qual',
        'pass_number',
        'phase_bias_ref_surface',
        'polarization_karin',
        'pole_tide',
        'rad_cloud_liquid_water',
        'rad_surface_type_flag',
        'rad_tmb_187',
        'rad_tmb_238',
        'rad_tmb_340',
        'rad_water_vapor',
        'rad_wet_tropo_cor',
        'rain_flag',
        'rain_rate',
        'sc_altitude',
        'sc_pitch',
        'sc_roll',
        'sc_yaw',
        'sea_state_bias_cor',
        'sea_state_bias_cor_2',
        'sig0_cor_atmos_model',
        'sig0_cor_atmos_rad',
        'sig0_karin',
        'sig0_karin_2',
        'sig0_karin_qual',
        'sig0_karin_uncert',
        'simulated_error_baseline_dilation',
        'simulated_error_karin',
        'simulated_error_orbital',
        'simulated_error_phase',
        'simulated_error_roll',
        'simulated_error_timing',
        'simulated_error_troposphere',
        'simulated_true_ssh_karin',
        'solid_earth_tide',
        'ssha_karin',
        'ssha_karin_2',
        'ssha_karin_qual',
        'ssh_karin',
        'ssh_karin_2',
        'ssh_karin_uncert',
        'swh_karin',
        'swh_karin_qual',
        'swh_karin_uncert',
        'swh_model',
        'swh_sea_state_bias',
        'time',
        'time_tai',
        'velocity_heading',
        'wind_speed_karin',
        'wind_speed_karin_2',
        'wind_speed_model_u',
        'wind_speed_model_v',
        'wind_speed_rad',
        'x_factor',
    ]
    start = numpy.datetime64('2000-01-01', 'D')
    end = numpy.datetime64('2000-02-01', 'D')
    delta = numpy.timedelta64(1, 'D')

    root = '/zcollection'
    fs.mkdir(root)
    fs.open(fs.sep.join((root, '.zcollection')), 'w').close()

    expected = []
    for date in numpy.arange(start, end, delta):
        item = date.item()
        partition = fs.sep.join(
            (root, f'year={item.year}', f'month={item.month:02d}',
             f'day={item.day:02d}'))
        expected.append(partition)
        fs.mkdirs(partition)

        _ = {fs.mkdirs(fs.sep.join((partition, item))) for item in variables}

        _ = {
            fs.open(fs.sep.join((partition, item)), 'w').close()
            for item in ['.zattrs', '.zgroup', '.zmetadata']
        }

    partitioning = Date(('dates', ), 'D')
    assert expected == list(partitioning.list_partitions(fs, root))



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Collection of Zarr groups
=========================
"""
from __future__ import annotations

from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Union,
)
import datetime
import functools
import io
import itertools
import json
import logging
import os
import pathlib
import types

import dask.bag.core
import dask.distributed
import dask.utils
import fsspec
import numpy
import xarray
import zarr
import zarr.storage

from .. import (
    dask_utils,
    dataset,
    expression,
    fs_utils,
    merging,
    meta,
    partitioning,
    storage,
    sync,
)
from .callable_objects import MapCallable, PartitionCallable, UpdateCallable
from .detail import (
    PartitioningProperties,
    _insert,
    _load_and_apply_indexer,
    _load_dataset,
    _load_dataset_with_overlap,
    _wrap_update_func,
    _wrap_update_func_with_overlap,
    try_infer_callable,
)

#: Type of functions filtering the partitions.
PartitionFilterCallback = Callable[[Dict[str, int]], bool]

#: Type of argument to filter the partitions.
PartitionFilter = Optional[Union[str, PartitionFilterCallback]]

#: Indexer's type.
Indexer = Iterable[Tuple[Tuple[Tuple[str, int], ...], slice]]

#: Name of the directory storing the immutable dataset.
_IMMUTABLE = '.immutable'

#: Module logger.
_LOGGER = logging.getLogger(__name__)


def variables(
    metadata: meta.Dataset,
    selected_variables: Iterable[str] | None = None
) -> tuple[dataset.Variable, ...]:
    """Return the variables defined in the dataset.

    Args:
        selected_variables: The variables to return. If None, all the
            variables are returned.

    Returns:
        The variables defined in the dataset.
    """
    selected_variables = selected_variables or metadata.variables.keys()
    return tuple(
        dataset.Variable(
            v.name, numpy.ndarray((0, ) * len(v.dimensions), v.dtype),
            v.dimensions, v.attrs, v.compressor, v.fill_value, v.filters)
        for k, v in metadata.variables.items() if k in selected_variables)


def build_indexer_args(
    collection: Collection,
    filters: PartitionFilter,
    indexer: Indexer,
    partitions: Iterable[str] | None = None,
) -> Iterator[tuple[tuple[tuple[str, int], ...], list[slice]]]:
    """Build the arguments for the indexer.

    Args:
        collection: The collection to index.
        filters: The partition filters.
        indexer: The indexer.
        partitions: The partitions to index. If None, all the partitions
            are indexed.

    Returns:
        An iterator containing the arguments for the indexer.
    """
    # Build an indexer dictionary between the partition scheme and
    # indexer.
    indexers_map: dict[tuple[tuple[str, int], ...], list[slice]] = {}
    _ = {
        indexers_map.setdefault(  # type: ignore[func-returns-value]
            partition_scheme, []).append(indexer)
        for partition_scheme, indexer in indexer
    }
    # Filter the selected partitions
    partitions = partitions or collection.partitions(filters=filters)
    selected_partitions = set(indexers_map) & {
        collection.partitioning.parse(item)
        for item in partitions
    }

    # For each provided partition scheme, retrieves the corresponding
    # indexer.
    return ((item, indexers_map[item]) for item in sorted(selected_partitions))


def _immutable_path(
    ds: meta.Dataset,
    partition_properties: PartitioningProperties,
) -> str | None:
    """Return the immutable path of the dataset.

    Args:
        ds: The dataset to process.
        partition_properties: The partitioning properties.

    Returns:
        The immutable path of the dataset containing data that are immutable
        relative to the partitioning or None if the dataset does not contain
        immutable data.
    """
    return fs_utils.join_path(
        partition_properties.dir, _IMMUTABLE) if ds.select_variables_by_dims(
            (partition_properties.dim, ), predicate=False) else None


def _write_immutable_dataset(ds: dataset.Dataset, axis: str, path: str,
                             fs: fsspec.AbstractFileSystem) -> None:
    """Write the immutable dataset.

    Args:
        ds: The dataset to write.
        axis: The partitioning axis.
        path: The path to the immutable dataset.
        fs: The file system that the partition is stored on.
        synchronizer: The synchronizer to use.
    """
    immutable_dataset = ds.select_variables_by_dims((axis, ), predicate=False)
    assert len(immutable_dataset.variables) != 0, (
        'The dataset to insert does not contain any variable '
        'that is not split.')
    _LOGGER.info('Creating the immutable dataset: %s', path)
    storage.write_zarr_group(immutable_dataset,
                             path,
                             fs,
                             sync.NoSync(),
                             parallel=False)


class Collection:
    """This class manages a collection of files in Zarr format stored in a set
    of subdirectories. These subdirectories split the data, by cycles or dates
    for example, in order to optimize access and updates, deletion or addition
    of new data.

    Args:
        axis: The axis of the collection.
        ds: The dataset containing the collection.
        partition_handler: The partitioning strategy for the collection.
        partition_base_dir: The base directory for the collection.
        mode: The mode of the collection.
        filesystem: The filesystem to use for the collection.
        synchronizer: The synchronizer to use for the collection.

    Raises:
        ValueError: If the axis does not exist in the dataset, if the
            partition key is not defined in the dataset or if the access mode
            is not supported.

    .. note::

        On the first call to the constructor, the dataset given as an argument
        is used to create the metadata of the collection. This information is
        used to validate datasets that are inserted in the collection.

        Normally, this class is not instantiated directly but through the
        :py:meth:`create_collection <zcollection.create_collection>` and
        :py:meth:`open_collection <zcollection.open_collection>` methods of this
        library.
    """
    #: Configuration filename of the collection.
    CONFIG: ClassVar[str] = '.zcollection'

    def __init__(
        self,
        axis: str,
        ds: meta.Dataset,
        partition_handler: partitioning.Partitioning,
        partition_base_dir: str,
        *,
        mode: str | None = None,
        filesystem: fsspec.AbstractFileSystem | str | None = None,
        synchronizer: sync.Sync | None = None,
    ) -> None:
        if axis not in ds.variables:
            raise ValueError(
                f'The variable {axis!r} is not defined in the dataset.')

        for varname in partition_handler.variables:
            if varname not in ds.variables:
                raise ValueError(
                    f'The partitioning key {varname!r} is not defined in '
                    'the dataset.')

        mode = mode or 'w'
        if mode not in ('r', 'w'):
            raise ValueError(f'The mode {mode!r} is not supported.')

        #: The axis of the collection.
        self.axis = axis
        #: The metadata that describes the dataset handled by the collection.
        self.metadata = ds
        #: The file system used to read/write the collection.
        self.fs = fs_utils.get_fs(filesystem)
        #: The partitioning strategy used to split the data.
        self.partitioning = partition_handler
        #: The partitioning properties (base directory and dimension).
        self.partition_properties = PartitioningProperties(
            fs_utils.normalize_path(self.fs, partition_base_dir),
            ds.variables[axis].dimensions[0],
        )
        #: The access mode of the collection.
        self.mode = mode
        #: The synchronizer used to synchronize the modifications.
        self.synchronizer = synchronizer or sync.NoSync()
        #: The path to the dataset that contains the immutable data relative
        #: to the partitioning.
        self._immutable = _immutable_path(ds, self.partition_properties)

        self._write_config(skip_if_exists=True)

        if mode == 'r':
            # pylint: disable=method-hidden
            # These methods are overloaded when the collection is opened in
            # readonly.
            for item in [
                    'add_variable',
                    'drop_partitions',
                    'drop_variable',
                    'insert',
                    'update',
            ]:
                assert hasattr(self, item), f'{item} is not a known method.'
                setattr(
                    self, item,
                    types.MethodType(Collection._unsupported_operation, self))
            # pylint: enable=method-hidden

    @property
    def immutable(self) -> bool:
        """Return True if the collection contains immutable data relative to
        the partitioning."""
        return self._immutable is not None

    def __str__(self) -> str:
        return (f'{self.__class__.__name__}'
                f'<filesystem={self.fs.__class__.__name__!r}, '
                f'partition_base_dir={self.partition_properties.dir!r}, '
                f'mode={self.mode!r}>')

    @staticmethod
    def _unsupported_operation(*args, **kwargs):
        """Raise an exception if the operation is not supported."""
        raise io.UnsupportedOperation('not writable')

    @classmethod
    def _config(cls, partition_base_dir: str) -> str:
        """Return the configuration path."""
        return fs_utils.join_path(partition_base_dir, cls.CONFIG)

    def _write_config(self, skip_if_exists: bool = False) -> None:
        """Write the configuration file."""
        base_dir = self.partition_properties.dir
        config = self._config(base_dir)
        exists = self.fs.exists(config)

        if skip_if_exists and exists:
            return

        message = ("Updating collection's configuration: %s"
                   if exists else 'Creating the collection: %s')
        _LOGGER.info(message, config)

        self.fs.makedirs(base_dir, exist_ok=True)

        params = {
            'axis': self.axis,
            'dataset': self.metadata.get_config(),
            'partitioning': self.partitioning.get_config(),
        }

        with self.fs.open(config, mode='w') as stream:
            json.dump(params, stream, indent=4)  # type: ignore[arg-type]

    def is_readonly(self) -> bool:
        """Return True if the collection is read-only."""
        return self.mode == 'r'

    @classmethod
    def from_config(
        cls,
        path: str,
        *,
        mode: str | None = None,
        filesystem: fsspec.AbstractFileSystem | str | None = None,
        synchronizer: sync.Sync | None = None,
    ) -> Collection:
        """Open a Collection described by a configuration file.

        Args:
            path: The path to the configuration file.
            mode: The mode of the collection.
            filesystem: The filesystem to use for the collection.
            synchronizer: The synchronizer to use for the collection.

        Returns:
            The collection.

        Raises:
            ValueError: If the provided directory does not contain a collection.
        """
        _LOGGER.info('Opening collection: %r', path)
        fs = fs_utils.get_fs(filesystem)
        config = cls._config(path)
        if not fs.exists(config):
            raise ValueError(f'zarr collection not found at path {path!r}')
        with fs.open(config) as stream:
            data = json.load(stream)
        return Collection(
            data['axis'],
            meta.Dataset.from_config(data['dataset']),
            partitioning.get_codecs(data['partitioning']),
            path,
            mode=mode or 'r',
            filesystem=fs,
            synchronizer=synchronizer,
        )

    def _is_selected(
        self,
        partition: Sequence[str],
        expr: Callable[[dict[str, int]], bool] | None,
    ) -> bool:
        """Return whether the partition is selected.

        Args:
            partition: The partition to check.
            expr: The expression used to filter the partition.

        Returns:
            Whether the partition is selected.
        """
        if expr is not None:
            return expr(dict(self.partitioning.parse('/'.join(partition))))
        return True

    # pylint: disable=method-hidden
    def insert(
        self,
        ds: xarray.Dataset | dataset.Dataset,
        *,
        merge_callable: merging.MergeCallable | None = None,
        npartitions: int | None = None,
    ) -> None:
        """Insert a dataset into the collection.

        Args:
            ds: The dataset to insert.
            merge_callable: The function to use to merge the existing data set
                already stored in partitions with the new partitioned data. If
                None, the new partitioned data overwrites the existing
                partitioned data.
            npartitions: The maximum number of partitions to process in
                parallel. By default, all partitions are processed in parallel
                in a single task.

        .. warning::

            Each worker will process a set of independent partitions. However,
            be careful, two different partitions can use the same file (chunk),
            therefore, the library that handles the storage of Dask arrays
            (HDF5, NetCDF, Zarr, etc.) must be compatible with concurrent
            access.

        Raises:
            ValueError: If the dataset mismatched the definition of the
                collection.
        """
        # pylint: disable=method-hidden
        if isinstance(ds, xarray.Dataset):
            ds = dataset.Dataset.from_xarray(ds)

        _LOGGER.info('Inserting of a %s dataset in the collection',
                     dask.utils.format_bytes(ds.nbytes))

        missing_variables = self.metadata.missing_variables(ds.metadata())
        for item in missing_variables:
            variable = self.metadata.variables[item]
            ds.add_variable(variable)

        client = dask_utils.get_client()

        # If the dataset contains variables that should not be partitioned.
        if self._immutable is not None:

            # On the first call, we store the immutable variables in a
            # a directory located at the root of the collection.
            if not self.fs.exists(self._immutable):
                _write_immutable_dataset(ds, self.axis, self._immutable,
                                         self.fs)

            # Remove the variables that should not be partitioned.
            ds = ds.select_variables_by_dims((self.axis, ))

        # Process the partitions to insert or update by batches to avoid
        # memory issues.
        partitions = tuple(
            self.partitioning.split_dataset(ds, self.partition_properties.dim))

        if npartitions is not None:
            if npartitions < 1:
                raise ValueError('The number of partitions must be positive')
            npartitions = len(partitions) // npartitions + 1

        scattared_ds = client.scatter(ds)
        for sequence in dask_utils.split_sequence(partitions, npartitions):
            futures = [
                dask_utils.simple_delayed('insert', _insert)(
                    partition,  # type: ignore[arg-type]
                    axis=self.axis,
                    ds=scattared_ds,
                    fs=self.fs,
                    merge_callable=merge_callable,
                    partitioning_properties=self.partition_properties,
                ) for partition in sequence
            ]
            client.compute(futures, sync=True)

    def _relative_path(self, path: str) -> str:
        """Return the relative path to the collection.

        Args:
            path: The path to the dataset.

        Returns:
            The relative path to the collection.
        """
        return pathlib.Path(path).relative_to(
            self.partition_properties.dir).as_posix()

    def partitions(
        self,
        *,
        filters: PartitionFilter = None,
        relative: bool = False,
    ) -> Iterator[str]:
        """List the partitions of the collection.

        Args:
            filters: The predicate used to filter the partitions to load. If
                the predicate is a string, it is a valid python expression to
                filter the partitions, using the partitioning scheme as
                variables. If the predicate is a function, it is a function
                that takes the partition scheme as input and returns a boolean.
            relative: Whether to return the relative path.

        Returns:
            The list of partitions.

        Example:
            >>> tuple(collection.partitions(
            ...     filters="year == 2019 and month == 1"))
            ('year=2019/month=01/day=01', 'year=2019/month=01/day=02/', ...)
            >>> tuple(collection.partitions(
            ...     filters=lambda x: x["year"] == 2019 and x["month"] == 1))
            ('year=2019/month=01/day=01', 'year=2019/month=01/day=02/', ...)
        """
        if isinstance(filters, str):
            expr: Any = expression.Expression(filters)
        elif callable(filters):
            expr = filters
        else:
            expr = None

        base_dir = self.partition_properties.dir
        sep = self.fs.sep

        for item in self.partitioning.list_partitions(self.fs, base_dir):
            if item == self._immutable:
                continue
            # Filtering on partition names
            partitions = item.replace(base_dir, '')
            entry = partitions.split(sep)

            if self._is_selected(entry, expr):
                yield self._relative_path(item) if relative else item

    # pylint: disable=method-hidden
    def drop_partitions(
        self,
        *,
        filters: PartitionFilter = None,
        timedelta: datetime.timedelta | None = None,
    ) -> None:
        # pylint: disable=method-hidden
        """Drop the selected partitions.

        Args:
            filters: The predicate used to filter the partitions to drop.
                To get more information on the predicate, see the
                documentation of the :meth:`partitions` method.
            timedelta: Select the partitions created before the
                specified time delta relative to the current time.

        Example:
            >>> collection.drop_partitions(filters="year == 2019")
            >>> collection.drop_partitions(
            ...     timedelta=datetime.timedelta(days=30))
        """
        now = datetime.datetime.now()
        client = dask_utils.get_client()
        folders = list(self.partitions(filters=filters))

        def is_created_before(path: str) -> bool:
            created = self.fs.created(path)
            return now - created > timedelta

        if timedelta is not None:
            folders = list(filter(is_created_before, folders))

        storage.execute_transaction(
            client, self.synchronizer,
            client.map(self.fs.rm, folders, recursive=True))

        def invalidate_cache(path):
            """Invalidate the cache."""
            _LOGGER.info('Dropped partition: %s', path)
            self.fs.invalidate_cache(path)

        tuple(map(invalidate_cache, folders))

    # pylint: disable=duplicate-code
    # false positive, no code duplication
    def map(
        self,
        func: MapCallable,
        *args,
        filters: PartitionFilter = None,
        partition_size: int | None = None,
        npartitions: int | None = None,
        selected_variables: Sequence[str] | None = None,
        **kwargs,
    ) -> dask.bag.core.Bag:
        """Map a function over the partitions of the collection.

        Args:
            func: The function to apply to every partition of the collection.
            *args: The positional arguments to pass to the function.
            filters: The predicate used to filter the partitions to process.
                To get more information on the predicate, see the
                documentation of the :meth:`partitions` method.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
            selected_variables: A list of variables to retain from the
                collection. If None, all variables are kept.
            **kwargs: The keyword arguments to pass to the function.

        Returns:
            A bag containing the tuple of the partition scheme and the result
            of the function.

        Example:
            >>> futures = collection.map(
            ...     lambda x: (x["var1"] + x["var2"]).values)
            >>> for item in futures:
            ...     print(item)
            [1.0, 2.0, 3.0, 4.0]
            [5.0, 6.0, 7.0, 8.0]
        """

        def _wrap(
            partition: str,
            func: PartitionCallable,
            selected_variables: Sequence[str] | None,
            *args,
            **kwargs,
        ) -> tuple[tuple[tuple[str, int], ...], Any]:
            """Wraps the function to apply on the partition.

            Args:
                func: The function to apply.
                partition: The partition to apply the function on.
                selected_variables: The list of variables to retain from the
                    partition.
                *args: The positional arguments to pass to the function.
                **kwargs: The keyword arguments to pass to the function.

            Returns:
                The result of the function.
            """
            ds = _load_dataset(self.fs, self._immutable, partition,
                               selected_variables)
            return self.partitioning.parse(partition), func(
                ds, *args, **kwargs)

        if not callable(func):
            raise TypeError('func must be a callable')

        bag = dask.bag.core.from_sequence(self.partitions(filters=filters),
                                          partition_size=partition_size,
                                          npartitions=npartitions)
        return bag.map(_wrap, func, selected_variables, *args, **kwargs)
        # pylint: enable=duplicate-code

    def map_overlap(
        self,
        func: MapCallable,
        depth: int,
        *args,
        filters: PartitionFilter = None,
        partition_size: int | None = None,
        npartition: int | None = None,
        selected_variables: Sequence[str] | None = None,
        **kwargs,
    ) -> dask.bag.core.Bag:
        """Map a function over the partitions of the collection with some
        overlap.

        Args:
            func: The function to apply to every partition of the collection.
                If ``func`` accepts a partition_info as a keyword
                argument, it will be passed a tuple with the name of the
                partitioned dimension and the slice allowing getting in the
                dataset the selected partition without the overlap.
            depth: The depth of the overlap between the partitions.
            *args: The positional arguments to pass to the function.
            filters: The predicate used to filter the partitions to process.
                To get more information on the predicate, see the
                documentation of the :meth:`partitions` method.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
            selected_variables: A list of variables to retain from the
                collection. If None, all variables are kept.
            **kwargs: The keyword arguments to pass to the function.

        Returns:
            A bag containing the tuple of the partition scheme and the result
            of the function.

        Example:
            >>> futures = collection.map_overlap(
            ...     lambda x: (x["var1"] + x["var2"]).values,
            ...     depth=1)
            >>> for item in futures:
            ...     print(item)
            [1.0, 2.0, 3.0, 4.0]
            [5.0, 6.0, 7.0, 8.0]
        """
        if not callable(func):
            raise TypeError('func must be a callable')

        add_partition_info = dask.utils.has_keyword(func, 'partition_info')

        def _wrap(
            partition: str,
            func: PartitionCallable,
            partitions: tuple[str, ...],
            selected_variables: Sequence[str] | None,
            depth: int,
            *args,
            **kwargs,
        ) -> tuple[tuple[tuple[str, int], ...], Any]:
            """Wraps the function to apply on the partition.

            Args:
                partition: The partition to apply the function on.
                func: The function to apply.
                partitions: The partitions to apply the function on.
                selected_variables: The list of variables to retain from the
                    partition.
                depth: The depth of the overlap between the partitions.
                *args: The positional arguments to pass to the function.
                **kwargs: The keyword arguments to pass to the function.

            Returns:
                The result of the function.
            """
            ds, indices = _load_dataset_with_overlap(
                depth, self.partition_properties.dim, self.fs, self._immutable,
                partition, partitions, selected_variables)

            if add_partition_info:
                kwargs = kwargs.copy()
                kwargs['partition_info'] = (self.partition_properties.dim,
                                            indices)

            # Finally, apply the function.
            return (self.partitioning.parse(partition),
                    func(ds, *args, **kwargs))

        partitions = tuple(self.partitions(filters=filters))
        bag = dask.bag.core.from_sequence(partitions,
                                          partition_size=partition_size,
                                          npartitions=npartition)
        return bag.map(_wrap, func, partitions, selected_variables, depth,
                       *args, **kwargs)

    def load(
        self,
        *,
        filters: PartitionFilter = None,
        indexer: Indexer | None = None,
        selected_variables: Iterable[str] | None = None,
    ) -> dataset.Dataset | None:
        """Load the selected partitions.

        Args:
            filters: The predicate used to filter the partitions to load.
                To get more information on the predicate, see the
                documentation of the :meth:`partitions` method.
            indexer: The indexer to apply.
            selected_variables: A list of variables to retain from the
                collection. If None, all variables are kept.

        Returns:
            The dataset containing the selected partitions.

        .. warning::

            If you select variables to load from the collection, do not insert
            the returned dataset otherwise all skipped variables will be reset
            with fill values.

        Example:
            >>> collection = ...
            >>> collection.load(
            ...     filters="year == 2019 and month == 3 and day % 2 == 0")
            >>> collection.load(
            ...     filters=lambda keys: keys["year"] == 2019 and
            ...     keys["month"] == 3 and keys["day"] % 2 == 0)
        """
        client = dask_utils.get_client()
        arrays: list[dataset.Dataset] = []
        if indexer is None:
            selected_partitions = tuple(self.partitions(filters=filters))
            if len(selected_partitions) == 0:
                return None

            # No indexer, so the dataset is loaded directly for each
            # selected partition.
            bag = dask.bag.core.from_sequence(
                self.partitions(filters=filters),
                npartitions=dask_utils.dask_workers(client, cores_only=True))
            arrays = bag.map(storage.open_zarr_group,
                             fs=self.fs,
                             selected_variables=selected_variables).compute()
        else:
            # Build the indexer arguments.
            args = tuple(build_indexer_args(self, filters, indexer))
            if len(args) == 0:
                return None

            bag = dask.bag.core.from_sequence(
                args,
                npartitions=dask_utils.dask_workers(client, cores_only=True))

            # Finally, load the selected partitions and apply the indexer.
            arrays = list(
                itertools.chain.from_iterable(
                    bag.map(
                        _load_and_apply_indexer,
                        fs=self.fs,
                        partition_handler=self.partitioning,
                        partition_properties=self.partition_properties,
                        selected_variables=selected_variables,
                    ).compute()))

        array = arrays.pop(0)
        if arrays:
            array = array.concat(arrays, self.partition_properties.dim)
        if self._immutable:
            array.merge(
                storage.open_zarr_group(self._immutable, self.fs,
                                        selected_variables))
        return array

    # pylint: disable=method-hidden
    def update(
        self,
        func: UpdateCallable,
        /,
        *args,
        depth: int = 0,
        filters: PartitionFilter | None = None,
        partition_size: int | None = None,
        selected_variables: Iterable[str] | None = None,
        **kwargs,
    ) -> None:
        # pylint: disable=method-hidden
        """Update the selected partitions.

        Args:
            func: The function to apply on each partition.
            *args: The positional arguments to pass to the function.
            depth: The depth of the overlap between the partitions. Default is
                0 (no overlap). If depth is greater than 0, the function is
                applied on the partition and its neighbors selected by the
                depth. If ``func`` accepts a partition_info as a keyword
                argument, it will be passed a tuple with the name of the
                partitioned dimension and the slice allowing getting in the
                dataset the selected partition.
            filters: The expression used to filter the partitions to update.
            partition_size: The number of partitions to update in a single
                batch. By default 1, which is the same as to map the function to
                each partition. Otherwise, the function is called on a batch of
                partitions.
            selected_variables: A list of variables to load from the collection.
                If None, all variables are loaded.
            **kwargs: The keyword arguments to pass to the function.

        Example:
            >>> import dask.array
            >>> import zcollection
            >>> def ones(ds):
            ...     return dict(var2=ds.variables["var1"].values * 0 + 1)
            >>> collection = zcollection.Collection("my_collection", mode="w")
            >>> collection.update(ones)
        """
        if not callable(func):
            raise TypeError('func must be a callable')

        try:
            one_partition = next(self.partitions(filters=filters))
        except StopIteration:
            return
        func_result = try_infer_callable(
            func,
            storage.open_zarr_group(one_partition, self.fs,
                                    selected_variables),
            self.partition_properties.dim, *args, **kwargs)
        unknown_variables = set(func_result) - set(
            self.metadata.variables.keys())
        if len(unknown_variables):
            raise ValueError(f'Unknown variables: {unknown_variables}')

        if depth == 0:
            local_func = _wrap_update_func(func, self.fs, self._immutable,
                                           selected_variables, *args, **kwargs)
        else:
            local_func = _wrap_update_func_with_overlap(
                depth, self.partition_properties.dim, func, self.fs,
                self._immutable, selected_variables, *args, **kwargs)

        _LOGGER.info('Updating of the (%s) variable in the collection',
                     ', '.join(repr(item) for item in func_result))
        client = dask_utils.get_client()
        batchs = dask_utils.split_sequence(
            tuple(self.partitions(filters=filters)), partition_size
            or dask_utils.dask_workers(client, cores_only=True))
        awaitables = client.map(local_func, tuple(batchs), key=func.__name__)
        storage.execute_transaction(client, self.synchronizer, awaitables)

    def _bag_from_partitions(
        self,
        filters: PartitionFilter | None = None,
    ) -> dask.bag.core.Bag:
        """Return a dask bag from the partitions.

        Args:
            filters: The predicate used to filter the partitions to load.

        Returns:
            The dask bag.
        """
        partitions = [*self.partitions(filters=filters)]
        return dask.bag.core.from_sequence(seq=partitions,
                                           npartitions=len(partitions))

    def drop_variable(
        self,
        variable: str,
    ) -> None:
        """Delete the variable from the collection.

        Args:
            variable: The variable to delete.

        Raises:
            ValueError: If the variable doesn't exist in the collection or is
                used by the partitioning.

        Example:
            >>> import zcollection
            >>> collection = zcollection.open_collection(
            ...     "my_collection", mode="w")
            >>> collection.drop_variable("my_variable")
        """
        _LOGGER.info('Dropping of the %r variable in the collection', variable)
        if variable in self.partitioning.variables:
            raise ValueError(
                f'The variable {variable!r} is part of the partitioning.')
        if variable not in self.metadata.variables:
            raise ValueError(
                f'The variable {variable!r} does not exist in the collection.')
        if self._immutable:
            ds = storage.open_zarr_group(self._immutable, self.fs)
            if variable in ds.variables:
                raise ValueError(
                    f'The variable {variable!r} is part of the immutable '
                    'dataset.')
        bag = self._bag_from_partitions()
        bag.map(storage.del_zarr_array, variable, self.fs).compute()
        del self.metadata.variables[variable]
        self._write_config()

    def add_variable(
        self,
        variable: meta.Variable | dataset.Variable,
    ) -> None:
        """Add a variable to the collection.

        Args:
            variable: The variable to add.

        Raises:
            ValueError: if the variable is already part of the collection, it
                doesn't use the partitioning dimension or use a dimension that
                is not part of the dataset.

        Example:
            >>> import zcollection
            >>> collection = zcollection.open_collection(
            ...     "my_collection", mode="w")
            >>> new_variable = meta.Variable(
            ...     name="my_variable",
            ...     dtype=numpy.dtype("int16"),
            ...     dimensions=("num_lines", "num_pixels"),
            ...     fill_value=32267,
            ...     attrs=(dataset.Attribute(name="my_attr", value=0), ),
            ... )
            >>> collection.add_variable(new_variable)
        """
        variable = dataset.get_variable_metadata(variable)
        _LOGGER.info('Adding of the %r variable in the collection',
                     variable.name)
        if self.partition_properties.dim not in variable.dimensions:
            raise ValueError(
                'The new variable must use the partitioning axis.')
        self.metadata.add_variable(variable)
        self._write_config()

        template = self.metadata.search_same_dimensions_as(variable)
        try:
            bag = self._bag_from_partitions()
            bag.map(storage.add_zarr_array, variable, template.name,
                    self.fs).compute()
        except Exception:
            self.drop_variable(variable.name)
            raise

    def iterate_on_records(
        self,
        *,
        relative: bool = False,
    ) -> Iterator[tuple[str, zarr.Group]]:
        """Iterate over the partitions and the zarr groups.

        Args:
            relative: If True, the paths are relative to the base directory.

        Returns
            The iterator over the partitions and the zarr groups.
        """
        yield from (
            (
                self._relative_path(item) if relative else item,
                zarr.open_consolidated(
                    self.fs.get_mapper(item),  # type: ignore
                    mode='r',
                )) for item in self.partitions())

    def variables(
        self,
        selected_variables: Iterable[str] | None = None
    ) -> tuple[dataset.Variable, ...]:
        """Return the variables of the collection.

        Args:
            selected_variables: The variables to return. If None, all the
                variables are returned.

        Returns:
            The variables of the collection.
        """
        return variables(self.metadata, selected_variables)

    def copy(
        self,
        target: str,
        *,
        filters: PartitionFilter | None = None,
        filesystem: fsspec.AbstractFileSystem | None = None,
        mode: str = 'w',
        npartitions: int | None = None,
        synchronizer: sync.Sync | None = None,
    ) -> Collection:
        """Copy the collection to a new location.

        Args:
            target: The target location.
            filters: The predicate used to filter the partitions to copy.
            filesystem: The file system to use. If None, the file system of the
                collection is used.
            mode: The mode used to open the collection copied. Default is 'w'.
            npartitions: The number of partitions top copy in parallel. Default
                is number of cores.
            synchronizer: The synchronizer used to synchronize the collection
                copied. Default is None.

        Returns:
            The new collection.

        Example:
            >>> import zcollection
            >>> collection = zcollection.open_collection(
            ...     "my_collection", mode="r")
            >>> collection.copy(target="my_new_collection")
        """
        _LOGGER.info('Copying of the collection to %r', target)
        if filesystem is None:
            filesystem = fs_utils.get_fs(target)
        client = dask_utils.get_client()
        npartitions = npartitions or dask_utils.dask_workers(client,
                                                             cores_only=True)

        # Sequence of (source, target) to copy split in npartitions
        args = tuple(
            dask_utils.split_sequence(
                [(item,
                  fs_utils.join_path(
                      target,
                      os.path.relpath(item, self.partition_properties.dir)))
                 for item in self.partitions(filters=filters)], npartitions))
        # Copy the selected partitions
        partial = functools.partial(fs_utils.copy_tree,
                                    fs_source=self.fs,
                                    fs_target=filesystem)

        def worker_task(args: Sequence[tuple[str, str]]) -> None:
            """Function call on each worker to copy the partitions."""
            tuple(map(lambda arg: partial(*arg), args))

        client.gather(client.map(worker_task, args))
        # Then the remaining files in the root directory (config, metadata,
        # etc.)
        fs_utils.copy_files([
            item['name']
            for item in self.fs.listdir(self.partition_properties.dir,
                                        detail=True) if item['type'] == 'file'
        ], target, self.fs, filesystem)
        return Collection.from_config(target,
                                      mode=mode,
                                      filesystem=filesystem,
                                      synchronizer=synchronizer)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Synchronization of concurrent accesses
======================================
"""
import abc


class Sync(abc.ABC):  # pragma: no cover
    """Interface of the classes handling the synchronization of concurrent
    accesses."""

    @abc.abstractmethod
    def __enter__(self) -> bool:
        ...

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...


class NoSync(Sync):
    """This class is used when the user does not want to synchronize accesses
    to the collection, in other words, when there is no concurrency."""

    def __enter__(self) -> bool:
        return True

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """As this class does not perform any synchronization, this method has
        nothing to do."""



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Dataset
=======
"""
from __future__ import annotations

from typing import Any, Iterable, Mapping, OrderedDict, Sequence
import collections

import dask.array.core
import dask.array.creation
import dask.array.ma
import dask.array.rechunk
import dask.array.routines
import dask.array.wrap
import dask.base
import dask.threaded
import fsspec
import xarray

from . import meta, variable
from .compressed_array import CompressedArray
from .meta import Attribute
from .type_hints import ArrayLike, NDArray, NDMaskedArray
from .variable import (
    Variable,
    _attributes_repr,
    _calculate_column_width,
    _dimensions_repr,
    _new_variable,
    _pretty_print,
)


def _dask_repr(array: dask.array.core.Array) -> str:
    """Get the string representation of a dask array.

    Args:
        array: A dask array.

    Returns:
        The string representation of the dask array.
    """
    chunksize = tuple(item[0] for item in array.chunks)
    return f'dask.array<chunksize={chunksize}>'


def _dataset_repr(ds: Dataset) -> str:
    """Get the string representation of a dataset.

    Args:
        ds: A dataset.

    Returns:
        The string representation of the dataset.
    """
    # Dimensions
    dims_str = _dimensions_repr(ds.dimensions)
    lines = [
        f'<{ds.__module__}.{ds.__class__.__name__}>',
        f'  Dimensions: {dims_str}', 'Data variables:'
    ]
    # Variables
    if len(ds.variables) == 0:
        lines.append('    <empty>')
    else:
        width = _calculate_column_width(ds.variables)
        for name, var in ds.variables.items():
            dims_str = f"({', '.join(map(str, var.dimensions))} "
            name_str = f'    {name:<{width}s} {dims_str} {var.dtype}'
            lines.append(_pretty_print(f'{name_str}: {_dask_repr(var.array)}'))
    # Attributes
    if len(ds.attrs):
        lines.append('  Attributes:')
        lines += _attributes_repr(ds.attrs)

    return '\n'.join(lines)


def _duplicate(var: variable.Variable,
               data: dask.array.core.Array) -> variable.Variable:
    """Duplicate the variable with a new data.

    Args:
        var: Variable to duplicate.
        data: The new data to use

    Returns:
        The duplicated variable
    """
    # pylint: disable=protected-access
    # _new is a protected member of this class
    return _new_variable(var.name, data, var.dimensions, var.attrs,
                         var.compressor, var.fill_value, var.filters)
    # pylint: enable=protected-access


class Dataset:
    """Hold variables, dimensions, and attributes that together form a dataset.

    Attrs:
        variables: Dataset variables
        attrs: Dataset attributes

    Raises:
        ValueError: If the dataset contains variables with the same dimensions
            but with different values.
    """
    __slots__ = ('dimensions', 'variables', 'attrs')

    def __init__(self,
                 variables: Iterable[variable.Variable],
                 attrs: Sequence[Attribute] | None = None) -> None:
        #: The list of global attributes on this dataset
        self.attrs = tuple(attrs or [])
        #: Dataset contents as dict of
        #: :py:class:`Variable <zcollection.variable.Variable>` objects.
        self.variables = collections.OrderedDict(
            (item.name, item) for item in variables)
        #: A dictionary of dimension names and their index in the dataset
        self.dimensions: dict[str, int] = {}

        for var in self.variables.values():
            try:
                for ix, dim in enumerate(var.dimensions):
                    if dim not in self.dimensions:
                        self.dimensions[dim] = var.array.shape[ix]
                    elif self.dimensions[dim] != var.array.shape[ix]:
                        raise ValueError(
                            f'variable {var.name} has conflicting '
                            'dimensions')
            except IndexError as exc:
                raise ValueError(
                    f'variable {var.name} has missing dimensions') from exc

    def __len__(self) -> int:
        return len(self.variables)

    def __bool__(self) -> bool:
        return bool(self.variables)

    def __getitem__(self, name: str) -> variable.Variable:
        """Return a variable from the dataset.

        Args:
            name: The name of the variable to return

        Returns:
            The variable

        Raises:
            KeyError: If the variable is not found
        """
        return self.variables[name]

    def __getstate__(self) -> tuple[Any, ...]:
        return self.dimensions, self.variables, self.attrs

    def __setstate__(
        self, state: tuple[dict[str, int], OrderedDict[str, variable.Variable],
                           tuple[Attribute, ...]]
    ) -> None:
        self.dimensions, self.variables, self.attrs = state

    @property
    def nbytes(self) -> int:
        """Return the total number of bytes in the dataset.

        Returns:
            The total number of bytes in the dataset
        """
        return sum(item.nbytes for item in self.variables.values())

    def add_variable(self,
                     var: meta.Variable,
                     /,
                     data: ArrayLike[Any] | None = None):
        """Add a variable to the dataset.

        Args:
            var: The variable to add
            data: The data to add to the variable. If not provided, the variable
                will be created with the default fill value.

        Raises:
            ValueError: If the variable added has dimensions that conflict with
                existing dimensions, or if the variable has dimensions not
                defined in the dataset.
        """
        if set(var.dimensions) - set(self.dimensions):
            raise ValueError(f'variable {var.name} has dimensions '
                             f'{var.dimensions} that are not in the dataset')

        if data is None:
            shape = tuple(self.dimensions[dim] for dim in var.dimensions)
            data = dask.array.wrap.full(shape, var.fill_value, dtype=var.dtype)
        else:
            for dim, size in zip(var.dimensions, data.shape):
                if size != self.dimensions[dim]:
                    raise ValueError(
                        f'Conflicting sizes for dimension {dim!r}: '
                        f'length {self.dimensions[dim]} on the data but length '
                        f'{size} defined in dataset.')
        self.variables[var.name] = Variable(
            var.name,
            data,  # type: ignore[arg-type]
            var.dimensions,
            var.attrs,
            var.compressor,
            var.fill_value,
            var.filters,
        )

    def rename(self, names: Mapping[str, str]) -> None:
        """Rename variables in the dataset.

        Args:
            names: A mapping from old names to new names

        Raises:
            ValueError: If the new names conflict with existing names
        """
        for old, new in names.items():
            if new in self.variables:
                raise ValueError(f'{new} already exists in the dataset')
            self.variables[new] = self.variables.pop(old).rename(new)

    def drops_vars(self, names: str | Sequence[str]) -> None:
        """Drop variables from the dataset.

        Args:
            names: Variable names to drop.
        """
        if isinstance(names, str) or not isinstance(names, Iterable):
            names = [names]
        tuple(map(self.variables.pop, names))

    def select_vars(self, names: str | Sequence[str]) -> Dataset:
        """Return a new dataset containing only the selected variables.

        Args:
            names: Variable names to select.

        Returns:
            A new dataset containing only the selected variables.
        """
        if isinstance(names, str) or not isinstance(names, Iterable):
            names = [names]
        return Dataset(
            [self.variables[name] for name in names],
            self.attrs,
        )

    def metadata(self) -> meta.Dataset:
        """Get the dataset metadata.

        Returns:
            Dataset metadata
        """
        return meta.Dataset(dimensions=tuple(self.dimensions.keys()),
                            variables=tuple(
                                item.metadata()
                                for item in self.variables.values()),
                            attrs=self.attrs)

    @staticmethod
    def from_xarray(ds: xarray.Dataset) -> Dataset:
        """Create a new dataset from an xarray dataset.

        Args:
            ds: Dataset to convert.

        Returns:
            New dataset.
        """
        variables = [
            Variable(
                name,  # type: ignore[arg-type]
                array.data,  # type: ignore[arg-type]
                tuple(array.dims),  # type: ignore[arg-type]
                tuple(
                    Attribute(*attr)  # type: ignore[arg-type]
                    for attr in array.attrs.items()),
                array.encoding.get('compressor', None),
                array.encoding.get('_FillValue', None),
                array.encoding.get('filters', None))
            for name, array in ds.variables.items()
        ]

        return Dataset(
            variables=variables,
            attrs=tuple(
                Attribute(*item)  # type: ignore[arg-type]
                for item in ds.attrs.items()))

    def to_xarray(self, **kwargs) -> xarray.Dataset:
        """Convert the dataset to an xarray dataset.

        Args:
            **kwargs: Additional parameters are passed through the function
                :py:func:`xarray.conventions.decode_cf_variables`.

        Returns:
            Dataset as an xarray dataset.
        """
        data_vars = collections.OrderedDict(
            (name, var.to_xarray()) for name, var in self.variables.items())
        attrs = collections.OrderedDict(
            (item.name, item.value) for item in self.attrs)
        data_vars, attrs, coord_names = xarray.conventions.decode_cf_variables(
            data_vars, attrs, **kwargs)
        ds = xarray.Dataset(data_vars, attrs=attrs)
        ds = ds.set_coords(coord_names.intersection(data_vars))
        return ds

    def to_dict(self,
                variables: Sequence[str] | None = None,
                **kwargs) -> dict[str, NDArray | NDMaskedArray]:
        """Convert the dataset to a dictionary, between the variable names and
        their data.

        Args:
            **kwargs: Additional parameters are passed through the function
                :py:func:`dask.compute`.

        Returns:
            Dictionary of variables.
        """
        variables = variables or tuple(self.variables.keys())
        arrays = tuple((key, value.data)
                       for key, value in self.variables.items()
                       if key in variables)
        return dict(dask.base.compute(*arrays, **kwargs))

    def isel(self, slices: dict[str, Any]) -> Dataset:
        """Return a new dataset with each array indexed along the specified
        slices.

        Args:
            slices: Dictionary of dimension names and slices

        Returns:
            New dataset.
        """
        dims_invalid = set(slices) - set(self.dimensions)
        if dims_invalid:
            raise ValueError(
                f'Slices contain invalid dimension name(s): {dims_invalid}')
        default = slice(None)
        variables = [
            var.isel(tuple(slices.get(dim, default) for dim in var.dimensions))
            for var in self.variables.values()
        ]
        return Dataset(variables=variables, attrs=self.attrs)

    def delete(self, indexer: slice | Sequence[int], axis: str) -> Dataset:
        """Return a new dataset without the data selected by the provided
        indices.

        Args:
            indexer: Indices to remove along the specified axis.
            axis: The axis along which to delete the subarrays defined
                in the dataset.

        Returns:
            New dataset.
        """
        variables = [
            _duplicate(
                var,
                dask.array.routines.delete(var.array, indexer,
                                           var.dimensions.index(axis)))
            for var in self.variables.values()
        ]
        return Dataset(variables=variables, attrs=self.attrs)

    def compute(self, **kwargs) -> Dataset:
        """Compute the dataset variables.

        Args:
            **kwargs: Additional parameters are passed through to
                :py:func:`dask.array.compute`.

        Returns:
            New dataset.
        """
        arrays = tuple(item.array for item in self.variables.values())
        arrays = dask.base.compute(*arrays, **kwargs)

        # Don't use _duplicate here because we because we want to transform
        # the numpy arrays computed by dask into dask arrays
        variables = [
            self.variables[k].duplicate(array)
            for k, array in zip(self.variables, arrays)
        ]
        return Dataset(variables=variables, attrs=self.attrs)

    def rechunk(self, **kwargs) -> Dataset:
        """Rechunk the dataset.

        Args:
            **kwargs: Keyword arguments are passed through to
                :py:func:`dask.array.rechunk.rechunk`.

        Returns:
            New dataset.

        .. seealso:: :py:func:`dask.array.rechunk`
        """
        variables = [var.rechunk(**kwargs) for var in self.variables.values()]
        return Dataset(variables=variables, attrs=self.attrs)

    def persist(
        self,
        *,
        compress: bool = False,
        **kwargs,
    ) -> Dataset:
        """Persist the dataset variables.

        Args:
            compress: If true, compress the data loaded into memory.
            **kwargs: Additional parameters are passed to the function
                :py:func:`dask.array.Array.persist`.

        Returns:
            The dataset with the variables persisted into memory.
        """
        if compress:
            for var in self.variables.values():
                var.array = var.array.map_blocks(CompressedArray,
                                                 fill_value=var.fill_value)
        arrays = dask.base.persist(
            *tuple(item.data for item in self.variables.values()), **kwargs)
        variables = self.variables
        for name, array in zip(self.variables, arrays):
            variables[name].array = array

        return self

    def concat(self, other: Dataset | Iterable[Dataset], dim: str) -> Dataset:
        """Concatenate datasets along a dimension.

        Args:
            other: Datasets to concatenate.
            dim: Dimension along which to concatenate the datasets.

        Returns:
            New dataset.

        Raises:
            ValueError: If the provided sequence of datasets is empty.
        """
        variables = []
        if not isinstance(other, Iterable):
            other = [other]
        if not other:
            raise ValueError('cannot concatenate an empty sequence')
        variables = [
            var.concat(tuple(item.variables[name] for item in other), dim)
            for name, var in self.variables.items()
        ]
        return Dataset(variables=variables, attrs=self.attrs)

    def merge(self, other: Dataset) -> None:
        """Merge the provided dataset into this dataset.

        Args:
            other: Dataset to merge into this dataset.
        """
        # Merge the variables
        for name, var in other.variables.items():

            # It's impossible to merge a variable with itself.
            if name in self.variables:
                raise ValueError(f'variable {name} already exists')
            self.variables[name] = var

        # If the dataset has common dimensions, they must be identical.
        same_dims = set(self.dimensions) & set(other.dimensions)
        if same_dims and not all(self.dimensions[dim] == other.dimensions[dim]
                                 for dim in same_dims):
            raise ValueError(f'dimensions {same_dims} are not identical')

        # Merge the dimensions.
        self.dimensions.update(other.dimensions)

        # Merge the attributes (overwriting any existing attributes).
        if self.attrs is None:
            self.attrs = other.attrs
        elif other.attrs is not None:
            attrs = dict(item.get_config() for item in self.attrs)
            attrs.update(item.get_config() for item in other.attrs)
            self.attrs = tuple(Attribute(*item) for item in attrs.items())

    def select_variables_by_dims(self,
                                 dims: Sequence[str],
                                 predicate: bool = True) -> Dataset:
        """Return a new dataset with only the variables that have the specified
        dimensions if predicate is true, otherwise return a new dataset with
        only the variables that do not have the specified dimensions.

        Args:
            dims: Dimensions to select.
            predicate: If true, select variables with the specified dimensions,
                otherwise select variables without the specified dimensions.

        Returns:
            New dataset or None if no variables match the predicate.
        """

        def _predicate_for_dimension_less(var: variable.Variable) -> bool:
            """Return true if the variable is selected by the predicate."""
            return (len(var.dimensions) == 0) == predicate

        def _predicate_for_dimension(var: variable.Variable) -> bool:
            """Return true if the variable is selected by the predicate."""
            return bool(set(var.dimensions) & set_of_dims) == predicate

        condition = (_predicate_for_dimension_less
                     if not dims else _predicate_for_dimension)

        set_of_dims = set(dims)
        variables = [var for var in self.variables.values() if condition(var)]
        return Dataset(variables=variables, attrs=self.attrs)

    def to_zarr(self,
                path: str,
                fs: fsspec.AbstractFileSystem | None = None,
                parallel: bool = True) -> None:
        """Write the dataset to a Zarr store.

        Args:
            path: Path to the Zarr store.
            fs: Filesystem to use.
            parallel: If true, write the data in parallel.
        """
        # pylint: disable=import-outside-toplevel, import-error
        # Avoid circular import
        import storage
        import sync

        # pylint: enable=import-outside-toplevel, import-error
        storage.write_zarr_group(self, path, fs or fsspec.filesystem('file'),
                                 sync.NoSync(), parallel)

    def __str__(self) -> str:
        return _dataset_repr(self)

    def __repr__(self) -> str:
        return _dataset_repr(self)


def get_variable_metadata(
        var: variable.Variable | meta.Variable) -> meta.Variable:
    """Get the variable metadata.

    Args:
        var: Variable to get the metadata for.

    Returns:
        Variable metadata.
    """
    if isinstance(var, Variable):
        return var.metadata()
    return var



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing variables
=================
"""
import pickle

import dask.array.core
import dask.array.ma
import numpy
import pytest
import zarr

from .. import meta, variable
# pylint: disable=unused-import # Need to import for fixtures
from .cluster import dask_client, dask_cluster

# pylint enable=unused-import


def test_maybe_truncate():
    """Test the truncation of a string to a given length."""
    data = list(range(1000))
    # pylint: disable=protected-access
    assert variable._maybe_truncate(data, 10) == '[0, 1, ...'
    assert variable._maybe_truncate(data, len(str(data))) == str(data)
    # pylint: enable=protected-access


def test_variable_masked_array(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test masked array."""
    var = create_test_variable()
    var2 = var.rename('var2')
    assert var2.array is var.array
    assert var2.name == 'var2'
    assert var2.dimensions == var.dimensions
    assert var2.attrs == var.attrs
    assert var2.compressor == var.compressor
    assert var2.filters == var.filters
    assert var2.fill_value == var.fill_value
    assert var2.dtype == var.dtype
    assert var2.shape == var.shape
    assert var2.size == var.size
    assert var2.ndim == var.ndim


def test_variable_not_equal():
    """Test if two values are different."""
    assert variable._not_equal(1, 2) is True
    assert variable._not_equal(1, 1) is False
    assert variable._not_equal(1, '1') is True
    assert variable._not_equal(1, numpy.nan) is True
    assert variable._not_equal(numpy.nan, numpy.nan) is False
    assert variable._not_equal(numpy.nan, 1) is True
    assert variable._not_equal(numpy.datetime64('NaT'),
                               numpy.datetime64('NaT')) is False
    assert variable._not_equal(numpy.datetime64('NaT'), 1) is True


def test_variable_as_asarray(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test converting array like to a dask array."""
    arr = numpy.arange(10)
    da, fill_value = variable._asarray(arr)
    assert isinstance(da, dask.array.core.Array)
    assert fill_value is None

    arr = numpy.ma.masked_equal(arr, 5)
    da, fill_value = variable._asarray(arr)
    assert isinstance(da, dask.array.core.Array)
    assert fill_value == 5

    da, fill_value = variable._asarray(dask.array.ma.masked_equal(arr, 5))
    assert isinstance(da, dask.array.core.Array)
    assert fill_value == 5

    with pytest.raises(ValueError):
        variable._asarray(numpy.ma.masked_equal(arr, 5), fill_value=6)

    with pytest.raises(ValueError):
        variable._asarray(numpy.ma.masked_equal(
            numpy.arange(numpy.datetime64(0, 'Y'),
                         numpy.datetime64(10, 'Y'),
                         dtype='M8[Y]'), numpy.datetime64(5, 'Y')),
                          fill_value=numpy.datetime64('NaT'))

    da = variable._asarray(numpy.ma.masked_equal(
        numpy.arange(numpy.datetime64(0, 'Y'),
                     numpy.datetime64(10, 'Y'),
                     dtype='M8[Y]'), numpy.datetime64('NaT')),
                           fill_value=numpy.datetime64('NaT'))


def create_test_variable(name='var1', fill_value=0):
    """Create a test variable."""
    return variable.Variable(name=name,
                             data=numpy.arange(10,
                                               dtype='int64').reshape(5, 2),
                             dimensions=('x', 'y'),
                             attrs=(variable.Attribute(name='attr',
                                                       value=1), ),
                             compressor=zarr.Blosc(cname='zstd', clevel=1),
                             fill_value=fill_value,
                             filters=(zarr.Delta('int64', 'int32'),
                                      zarr.Delta('int32', 'int32')))


def test_variable(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test variable creation."""
    var = create_test_variable()
    assert var.name == 'var1'
    assert var.dtype == numpy.dtype('int64')
    assert var.shape == (5, 2)
    assert var.dimensions == ('x', 'y')
    assert var.attrs == (variable.Attribute(name='attr', value=1), )
    assert var.compressor.cname == 'zstd'  # type: ignore
    assert var.compressor.clevel == 1  # type: ignore
    assert var.fill_value == 0
    assert var.size == 10
    assert var.nbytes == 80
    assert var.filters == (
        zarr.Delta('int64', 'int32'),
        zarr.Delta('int32', 'int32'),
    )
    assert numpy.all(var.values == numpy.arange(10).reshape(5, 2))
    assert numpy.all(var.values == var.values)
    assert tuple(var.dimension_index()) == (('x', 0), ('y', 1))
    assert isinstance(var.metadata(), meta.Variable)
    assert isinstance(str(var), str)
    assert isinstance(repr(var), str)

    other = pickle.loads(pickle.dumps(var))
    assert other.name == 'var1'
    assert other.dtype == numpy.dtype('int64')
    assert other.shape == (5, 2)
    assert other.dimensions == ('x', 'y')
    assert other.attrs == (variable.Attribute(name='attr', value=1), )
    assert other.compressor.cname == 'zstd'  # type: ignore
    assert other.compressor.clevel == 1  # type: ignore
    assert other.fill_value == 0
    assert other.size == 10
    assert other.nbytes == 80
    assert other.filters == (
        zarr.Delta('int64', 'int32'),
        zarr.Delta('int32', 'int32'),
    )
    assert numpy.all(var.values == other.values)

    def foo(a, b):
        return a + b

    assert numpy.all(
        foo(var, var.values) == numpy.arange(10).reshape(5, 2) +
        numpy.arange(10).reshape(5, 2))

    assert numpy.all(
        foo(var, var.data).compute() == numpy.arange(10).reshape(5, 2) +
        numpy.arange(10).reshape(5, 2))

    var.data = numpy.ones((10, 4), dtype='int64')
    assert var.data.shape == (10, 4)
    assert isinstance(var.data, dask.array.core.Array)
    assert numpy.all(var.values == 1)

    with pytest.raises(ValueError):
        var.data = numpy.ones((10, 4, 2), dtype='int64')


def test_variable_duplicate(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test of the duplication of variables."""
    var = create_test_variable()
    other = var.duplicate(var.array * 2)
    assert other.name == 'var1'
    assert other.dtype == numpy.dtype('int64')
    assert other.shape == (5, 2)
    assert other.dimensions == ('x', 'y')
    assert other.attrs == (variable.Attribute(name='attr', value=1), )
    assert other.compressor.cname == 'zstd'  # type: ignore
    assert other.compressor.clevel == 1  # type: ignore
    assert other.fill_value == 0
    assert other.filters == (
        zarr.Delta('int64', 'int32'),
        zarr.Delta('int32', 'int32'),
    )
    assert numpy.all(var.values == other.values / 2)  # type: ignore
    assert var.metadata() == other.metadata()

    with pytest.raises(ValueError):
        var.duplicate(numpy.ones((10, 4, 2), dtype='int64'))


def test_variable_concat(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test concatenation of variables."""
    var_a = create_test_variable()
    var_b = create_test_variable()
    var_c = create_test_variable()

    vard = var_a.concat((var_b, var_c), 'x')
    assert numpy.all(vard.values == numpy.concatenate(
        (var_a.values, var_b.values, var_c.values), axis=0))

    vard = var_a.concat(var_b, 'x')
    assert numpy.all(
        vard.values == numpy.concatenate((var_a.values, var_b.values), axis=0))

    with pytest.raises(ValueError):
        var_a.concat([], 'y')


def test_variable_datetime64_to_xarray(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test conversion to xarray."""
    dates = numpy.arange(
        numpy.datetime64('2000-01-01', 'ms'),
        numpy.datetime64('2000-02-01', 'ms'),
        numpy.timedelta64('1', 'h'),
    )
    var = variable.Variable(
        name='time',
        data=dates,
        dimensions=('num_lines', ),
        attrs=(variable.Attribute(name='attr', value=1), ),
        compressor=zarr.Blosc(),
        filters=(zarr.Delta('int64', 'int64'), ),
    )
    xr_var = var.to_xarray()
    assert xr_var.dims == ('num_lines', )
    assert xr_var.attrs == dict(attr=1)
    assert xr_var.dtype == 'datetime64[ns]'


def test_variable_timedelta64_to_xarray(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test conversion to xarray."""
    delta = numpy.diff(
        numpy.arange(
            numpy.datetime64('2000-01-01', 'ms'),
            numpy.datetime64('2000-02-01', 'ms'),
            numpy.timedelta64('1', 'h'),
        ))

    var = variable.Variable(
        name='timedelta',
        data=delta,
        dimensions=('num_lines', ),
        attrs=(variable.Attribute(name='attr', value=1), ),
        compressor=zarr.Blosc(),
        filters=(zarr.Delta('int64', 'int64'), ),
    )
    xr_var = var.to_xarray()
    assert xr_var.dims == ('num_lines', )
    assert xr_var.attrs == dict(attr=1)
    assert xr_var.dtype.kind == 'm'


def test_variable_dimension_less(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Concatenate two dimensionless variables."""
    data = numpy.array([0, 1], dtype=numpy.int32)
    args = ('nv', data, ('nv', ), (variable.Attribute('comment', 'vertex'),
                                   variable.Attribute('units', '1')))
    n_vertex = variable.Variable(*args)
    assert n_vertex.fill_value is None
    metadata = n_vertex.metadata()
    assert metadata.fill_value is None
    assert meta.Variable.from_config(metadata.get_config()) == metadata

    other = variable.Variable(*args)

    concatenated = n_vertex.concat((other, ), 'time')
    assert numpy.all(concatenated.values == n_vertex.values)
    assert concatenated.metadata() == n_vertex.metadata()


def test_variable_getitem(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    var = create_test_variable()
    values = var.values
    result = var[0].compute()
    assert numpy.all(result == values[0])
    result = var[0:2].compute()
    assert numpy.all(result == values[0:2])
    result = var[0:2, 0].compute()
    assert numpy.all(result == values[0:2, 0])
    result = var[0:2, 0:2].compute()
    assert numpy.all(result == values[0:2, 0:2])


def test_variable_fill(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test filling of variables."""
    var = create_test_variable()
    assert not var.values.all() is numpy.ma.masked
    var.fill()
    assert var.values.all() is numpy.ma.masked


def test_variable_rechunk(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test rechunking of variables."""
    var = create_test_variable()
    values = var.values
    var = var.rechunk()
    assert numpy.all(var.values == values)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Configuration metadata
======================
"""
from __future__ import annotations

from typing import Any, Iterable, Sequence
import abc

import numcodecs.abc
import numpy
import zarr.codecs
import zarr.meta

from .type_hints import DTypeLike


class Pair(abc.ABC):
    """Handle pair key/value.

    Args:
        name: name of the key.
        value: value of the key.
    """
    __slots__ = ('name', 'value')

    def __init__(self, name: str, value: Any) -> None:
        #: Name of the key.
        self.name = name
        #: Value of the key.
        self.value = self._encode(value)

    @staticmethod
    def _encode(value: Any) -> Any:
        """Encode an attribute value as something that can be serialized as
        JSON."""
        if isinstance(value, numpy.ndarray):
            return value.tolist()

        if isinstance(value, numpy.generic):
            return value.item()

        return value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.name!r}, {self.value})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Pair):
            return False
        return self.get_config() == other.get_config()

    def get_config(self) -> tuple[str, Any]:
        """Get the key/value pair configuration."""
        return self.name, self.value

    @staticmethod
    @abc.abstractmethod
    def from_config(data: tuple[str, Any]) -> Pair:
        """Create a new Pair from the given key/value pair configuration."""


class Dimension(Pair):
    """Handle the metadata of a dataset dimension.

    Args:
        name: name of the dimension.
        value: value of the dimension.
    """

    @staticmethod
    def from_config(data: tuple[str, Any]) -> Dimension:
        """Creates a new instance from its metadata.

        Returns:
            Dimension: a new dimension.
        """
        return Dimension(*data)


class Attribute(Pair):
    """Handle the metadata of a dataset attribute.

    Args:
        name: name of the attribute.
        value: value of the attribute.
    """

    @staticmethod
    def from_config(data: tuple[str, Any]) -> Attribute:
        """Create a new instance from its metadata.

        Args:
            data: attribute configuration.

        Returns:
            Attribute: a new attribute.
        """
        return Attribute(*data)


class Variable:
    """Handle the metadata of a dataset variable.

    Args:
        name: name of the variable
        dtype: data type of the variable
        dimensions: names of the dimensions of the variable
        attrs: attributes of the variable
        compressor: compression codec for the variable
        fill_value: fill value for the variable
        filters: filters for the variable
    """
    __slots__ = ('attrs', 'compressor', 'dimensions', 'dtype', 'fill_value',
                 'filters', 'name')

    def __init__(self,
                 name: str,
                 dtype: DTypeLike,
                 dimensions: Sequence[str] | None = None,
                 attrs: Sequence[Attribute] | None = None,
                 compressor: numcodecs.abc.Codec | None = None,
                 fill_value: Any | None = None,
                 filters: Sequence[numcodecs.abc.Codec] | None = None) -> None:
        attrs = attrs or tuple()

        #: Attributes of the variable.
        self.attrs = tuple(attrs)
        #: Compression codec for the variable.
        self.compressor = compressor
        #: Dimensions of the variable.
        self.dimensions = tuple(dimensions or ())
        #: Data type of the variable.
        self.dtype = numpy.dtype(dtype)
        #: Fill value for the variable.
        self.fill_value = fill_value
        #: Filter codecs for the variable.
        self.filters = filters or tuple()
        #: Variable name.
        self.name = name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.name!r})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Variable):
            return False
        return self.get_config() == other.get_config()

    def get_config(self) -> dict[str, Any]:
        """Get the variable metadata.

        Returns:
            variable configuration.
        """
        compressor = self.compressor
        compressor = compressor.get_config(
        ) if compressor is not None else None
        filters = tuple(item.get_config() for item in self.filters)

        return {
            'attrs': sorted(item.get_config() for item in self.attrs),
            'compressor': compressor,
            'dimensions': self.dimensions,
            'dtype': zarr.meta.encode_dtype(self.dtype),
            'fill_value': zarr.meta.encode_fill_value(self.fill_value,
                                                      self.dtype),
            'filters': filters,
            'name': self.name,
        }

    @staticmethod
    def from_config(data: dict[str, Any]) -> Variable:
        """Create a new variable from the given variable configuration.

        Args:
            data: variable configuration.

        Returns:
            new variable.
        """

        def get_codec(codec) -> numcodecs.abc.Codec | None:
            """Get the codec from its configuration."""
            return zarr.codecs.get_codec(codec) if codec is not None else None

        dtype = zarr.meta.decode_dtype(data['dtype'])
        filters: Sequence[numcodecs.abc.Codec] = tuple(
            zarr.codecs.get_codec(item) for item in data['filters']
            if item is not None)

        return Variable(
            data['name'],
            dtype,
            data['dimensions'],
            tuple(Attribute.from_config(item) for item in data['attrs']),
            get_codec(data['compressor']),
            zarr.meta.decode_fill_value(data['fill_value'], dtype),
            filters,
        )


class Dataset:
    """Handle the metadata of a dataset.

    Args:
        dimensions: dimensions of the dataset
        variables: variables of the dataset
        attrs: attributes of the dataset
    """
    __slots__ = ('dimensions', 'variables', 'attrs')

    def __init__(self,
                 dimensions: Sequence[str],
                 variables: Sequence[Variable],
                 attrs: Sequence[Attribute] | None = None) -> None:
        #: Dimensions of the dataset.
        self.dimensions = tuple(dimensions)

        #: Variables of the dataset.
        self.variables = {item.name: item for item in variables}

        #: Attributes of the dataset.
        self.attrs = attrs or []

    def select_variables(
        self,
        keep_variables: Iterable[str] | None = None,
        drop_variables: Iterable[str] | None = None,
    ) -> set[str]:
        """Select variables to keep or drop from the dataset.

        Args:
            keep_variables: A list of variables to retain from the Dataset.
                If None, all variables are kept.
            drop_variables: A list of variables to exclude from the Dataset.
                If None, no variables are dropped.

        Returns:
            The selected variables.
        """
        result = set(self.variables)
        if keep_variables is not None:
            result &= set(keep_variables)
        if drop_variables is not None:
            result -= set(drop_variables)
        return result

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Dataset):
            return False
        return self.get_config() == other.get_config()

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def get_config(self) -> dict[str, Any]:
        """Get the dataset metadata.

        Returns:
            Dataset configuration.
        """
        return {
            'attrs':
            sorted(item.get_config() for item in self.attrs),
            'dimensions':
            self.dimensions,
            'variables':
            tuple(self.variables[name].get_config()
                  for name in sorted(self.variables)),
        }

    def add_variable(self, variable: Variable) -> None:
        """Add a variable to the dataset.

        Args:
            variable: variable to add.

        Raises:
            TypeError: if the variable is not a Variable.
            ValueError: if the variable already exists or if the variable
                dimensions don't match the dataset dimensions.
        """
        if not isinstance(variable, Variable):
            raise TypeError(
                f'variable must be a Variable, not {type(variable)}')
        if variable.name in self.variables:
            raise ValueError(
                f'The variable {variable.name!r} already exists in the '
                'collection.')
        dimensions = set(self.dimensions)
        # Looking for unknown dimensions.
        if (set(variable.dimensions) | dimensions) != dimensions:
            raise ValueError(
                'The new variable must use the dataset dimensions.')
        self.variables[variable.name] = variable

    @staticmethod
    def from_config(data: dict[str, Any]) -> Dataset:
        """Create a new dataset from the given dataset configuration.

        Args:
            data: dataset configuration.

        Returns:
            New dataset.
        """
        return Dataset(
            dimensions=data['dimensions'],
            variables=tuple(
                Variable.from_config(item) for item in data['variables']),
            attrs=tuple(Attribute.from_config(item) for item in data['attrs']),
        )

    def search_same_dimensions_as(self, variable: Variable) -> Variable:
        """Search for a variable in this dataset with the same dimensions as
        the given variable.

        Args:
            variable: The variable used for searching.

        Returns:
            The variable having the same dimensions as the supplied variable.

        Raises:
            ValueError: If no variable with the same dimensions as the given
                variable is found.
        """
        for item in self.variables.values():
            if item.dimensions == variable.dimensions:
                return item
        raise ValueError('No variable using the same dimensions exists.')

    def missing_variables(self, other: Dataset) -> Sequence[str]:
        """Finds the variables in the provided dataset that are not in this
        instance.

        Args:
            other: The dataset to compare against.

        Returns:
            The names of the missing variables in this dataset relative to the
            provided dataset.
        """
        this = set(self.variables)
        others = set(other.variables)

        if len(others - this):
            raise ValueError('The reference dataset does not define the '
                             f'{", ".join(others - this)} variables that are '
                             'defined in this dataset.')

        return tuple(this - others)

    def select_variables_by_dims(self,
                                 dims: Sequence[str],
                                 predicate: bool = True) -> set[str]:
        """Select variables that have at least one dimension in the given
        dimensions depending on the predicate.

        Args:
            dims: The dimensions to select.
            predicate: If True, select variables that have the given dimensions.
                If False, select variables that don't have the given dimensions.

        Returns:
            The names of the variables that have the given dimensions.
        """
        if len(dims) == 0:
            return {
                name
                for name, var in self.variables.items()
                if (len(var.dimensions) == 0) == predicate
            }

        set_of_dims = set(dims)
        return {
            name
            for name, var in self.variables.items()
            if bool(set(var.dimensions) & set_of_dims) == predicate
        }
